"""
This is a Python wrapper for the R package ['melsim'](https://github.com/sebsilas/melsim).
This wrapper allows the user to easily interface with the melsim package using the AMADS Score object.
Melsim is a package for computing similarity between melodies, and is being developed by
[Sebastian Silas](https://sebsilas.com/) and Klaus Frieler
([https://www.aesthetics.mpg.de/en/the-institute/people/klaus-frieler.html](
https://www.aesthetics.mpg.de/en/the-institute/people/klaus-frieler.html)).

Melsim is based on SIMILE, which was written by Daniel Müllensiefen and Klaus Frieler in 2003/2004.
This package is used to compare two or more melodies pairwise across a range of similarity measures.
Not all similarity measures are implemented in melsim, but the ones that are can be used here.
All of the following similarity measures are implemented and functional in melsim:
Please be aware that the names of the similarity measures are case-sensitive.

| Num | Name           |
|-----|----------------|
| 1   | Jaccard        |
| 2   | Kulczynski2    |
| 3   | Russel         |
| 4   | Faith          |
| 5   | Tanimoto       |
| 6   | Dice           |
| 7   | Mozley         |
| 8   | Ochiai         |
| 9   | Simpson        |
| 10  | cosine         |
| 11  | angular        |
| 12  | correlation    |
| 13  | Tschuprow      |
| 14  | Cramer         |
| 15  | Gower          |
| 16  | Euclidean      |
| 17  | Manhattan      |
| 18  | supremum       |
| 19  | Canberra       |
| 20  | Chord          |
| 21  | Geodesic       |
| 22  | Bray           |
| 23  | Soergel        |
| 24  | Podani         |
| 25  | Whittaker      |
| 26  | eJaccard       |
| 27  | eDice          |
| 28  | Bhjattacharyya |
| 29  | divergence     |
| 30  | Hellinger      |
| 31  | edit_sim_utf8  |
| 32  | edit_sim       |
| 33  | Levenshtein    |
| 34  | sim_NCD        |
| 35  | const          |
| 36  | sim_dtw        |



The following similarity measures are not currently functional in melsim:

| Num | Name | Type |
|-----|------|------|
| 1 | count_distinct | set-based |
| 2 | tversky | set-based |
| 3 | simple matching | |
| 4 | braun_blanquet | set-based |
| 5 | minkowski | vector-based |
| 6 | ukkon | distribution-based |
| 7 | sum_common | distribution-based |
| 8 | distr_sim | distribution-based |
| 9 | stringdot_utf8 | sequence-based |
| 10 | pmi | special |
| 11 | sim_emd | special |

Further to the similarity measures, melsim allows the user to specify which domain the
similarity should be calculated for. This is referred to as a “transformation” in melsim,
and all of the following transformations are implemented and functional:

| Num | Name |
|-----|------|
| 1 | pitch |
| 2 | int |
| 3 | fuzzy_int |
| 4 | parsons |
| 5 | pc |
| 6 | ioi_class |
| 7 | duration_class |
| 8 | int_X_ioi_class |
| 9 | implicit_harmonies |

The following transformations are not currently functional in melsim:

| Num | Name |
|-----|------|
| 1 | ioi |
| 2 | phrase_segmentation |
"""

import json
import math
import os
import shutil
import subprocess
from itertools import combinations
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from tenacity import RetryError, Retrying, stop_after_attempt, wait_exponential
from tqdm import tqdm

r_base_packages = ["base", "utils"]
r_cran_packages = [
    "tibble",
    "R6",
    "remotes",
    "dplyr",
    "magrittr",
    "proxy",
    "purrr",
    "purrrlyr",
    "tidyr",
    "yaml",
    "stringr",
    "emdist",
    "dtw",
    "ggplot2",
    "cba",
    "jsonlite",
]
r_github_packages = ["melsim"]
github_repos = {
    "melsim": "sebsilas/melsim",
}

# where to find Rscript (cached here after first lookup)
_rscript_path: Optional[str] = None


def run_script_in_r(script: str, text: bool = True) -> str:
    """Run an R script and return its output.

    Parameters
    ----------
    script : str
        R script to run

    Returns
    -------
    str
        Standard output from the R script

    Raises
    ------
    RuntimeError
        If there is an error running the R script
    """
    global _rscript_path
    if not _rscript_path:
        _rscript_path = _find_rscript()
    result = subprocess.run(
        [_rscript_path, "-e", script],
        capture_output=True,
        text=text,
        check=True,
    )
    return result.stdout.strip()


def _find_rscript() -> str:
    """Find the Rscript executable.

    Returns
    -------
    str
        Path to Rscript executable

    Raises
    ------
    RuntimeError
        If Rscript is not found
    """
    # Try common locations for Rscript
    rscript_path = shutil.which("Rscript")

    if rscript_path:
        return rscript_path

    # Try common macOS homebrew locations
    common_paths = [
        "/opt/homebrew/bin/Rscript",  # Apple Silicon homebrew
        "/usr/local/bin/Rscript",  # Intel homebrew
        "/Library/Frameworks/R.framework/Resources/bin/Rscript",  # R.app
    ]

    for path in common_paths:
        if os.path.exists(path) and os.access(path, os.X_OK):
            return path

    raise RuntimeError(
        "Rscript not found. Please install R and ensure Rscript is in your PATH. "
        "You can install R from https://cran.r-project.org/ or using Homebrew: brew install r"
    )


def check_r_packages_installed(
    install_missing: bool = False, n_retries: int = 3
):
    """Check if required R packages are installed.

    Parameters
    ----------
    install_missing : bool, default=False
        If True, attempt to install missing packages automatically.
    n_retries : int, default=3
        Number of retries for installing each missing package.

    Raises
    ------
    ImportError
        If required packages are missing and install_missing is False.
    RuntimeError
        If there is an error checking or installing packages.
    """
    # Create R script to check package installation using base R only
    check_script = """
    packages <- c({packages})
    missing <- packages[!sapply(packages, requireNamespace, quietly = TRUE)]
    if (length(missing) > 0) {{
        cat(paste0('"', missing, '"', collapse = ","))
    }} else {{
        cat("")
    }}
    """

    # Format package list
    packages_str = ", ".join(
        [f'"{p}"' for p in r_cran_packages + r_github_packages]
    )
    check_script = check_script.format(packages=packages_str)

    # Run R script
    try:
        output = run_script_in_r(check_script)

        # Parse the output - if empty, no missing packages
        if not output:
            missing_packages = []
        else:
            # Parse comma-separated quoted strings
            missing_packages = [pkg.strip('"') for pkg in output.split(",")]

        if missing_packages:
            if install_missing:
                for package in missing_packages:
                    try:
                        for attempt in Retrying(
                            stop=stop_after_attempt(n_retries),
                            wait=wait_exponential(multiplier=1, min=1, max=10),
                        ):
                            with attempt:
                                install_r_package(package)
                    except RetryError as e:
                        raise RuntimeError(
                            f"Failed to install R package '{package}' after {n_retries} attempts. "
                            "See above for the traceback."
                        ) from e
            else:
                raise ImportError(
                    f"Packages {missing_packages} are required but not installed. "
                    "You can install them by running: install_dependencies()"
                )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error checking R packages: {e.stderr}")


def install_r_package(package: str):
    """Install an R package.

    Parameters
    ----------
    package : str
        Name of the R package to install.

    Raises
    ------
    ValueError
        If the package type is unknown.
    """
    if package in r_cran_packages:
        print(f"Installing CRAN package '{package}'...")
        install_script = f"""
        utils::chooseCRANmirror(ind=1)
        utils::install.packages("{package}", dependencies=TRUE)
        """
        _ = run_script_in_r(install_script)
    elif package in r_github_packages:
        print(f"Installing GitHub package '{package}'...")
        repo = github_repos[package]
        install_script = f"""
        if (!requireNamespace("remotes", quietly = TRUE)) {{
            utils::install.packages("remotes")
        }}
        remotes::install_github("{repo}", upgrade="always", dependencies=TRUE)
        """
        _ = run_script_in_r(install_script)
    else:
        raise ValueError(f"Unknown package type for '{package}'")


def install_dependencies():
    """Install all required R packages.

    Raises
    ------
    ImportError
        If required packages are missing and install_missing is False.
    RuntimeError
        If there is an error checking or installing packages.
    """
    # Check which packages need to be installed using base R only
    check_script = """
    packages <- c({packages})
    missing <- packages[!sapply(packages, requireNamespace, quietly = TRUE)]
    if (length(missing) > 0) {{
        cat(paste0('"', missing, '"', collapse = ","))
    }} else {{
        cat("")
    }}
    """

    # Check CRAN packages
    packages_str = ", ".join([f'"{p}"' for p in r_cran_packages])
    check_script_cran = check_script.format(packages=packages_str)

    try:
        output = run_script_in_r(check_script_cran)

        # Parse the output - if empty, no missing packages
        if not output:
            missing_cran = []
        else:
            # Parse comma-separated quoted strings
            missing_cran = [pkg.strip('"') for pkg in output.split(",")]

        if missing_cran:
            print("Installing missing CRAN packages...")
            cran_script = f"""
            utils::chooseCRANmirror(ind=1)
            utils::install.packages(c({", ".join([f'"{p}"' for p in missing_cran])}), dependencies=TRUE)
            """
            _ = run_script_in_r(cran_script)
        else:
            print("Skipping install: All CRAN packages are already installed.")
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error checking CRAN packages: {e.stderr}")

    # Check GitHub packages
    packages_str = ", ".join([f'"{p}"' for p in r_github_packages])
    check_script_github = check_script.format(packages=packages_str)

    try:
        output = run_script_in_r(check_script_github)

        # Parse the output - if empty, no missing packages
        if not output:
            missing_github = []
        else:
            # Parse comma-separated quoted strings
            missing_github = [pkg.strip('"') for pkg in output.split(",")]

        if missing_github:
            print("Installing missing GitHub packages...")
            for package in missing_github:
                repo = github_repos[package]
                print(f"Installing {package} from {repo}...")
                install_script = f"""
                if (!requireNamespace("remotes", quietly = TRUE)) {{
                    utils::install.packages("remotes")
                }}
                remotes::install_github("{repo}", upgrade="always", dependencies=TRUE)
                """
                _ = run_script_in_r(install_script)
        else:
            print(
                "Skipping install: All GitHub packages are already installed."
            )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error checking GitHub packages: {e.stderr}")

    print("All dependencies are installed and up to date.")


def check_python_package_installed(package: str):
    """Check if a Python package is installed.

    Parameters
    ----------
    package : str
        Name of the Python package to check.

    Raises
    ------
    ImportError
        If the package is not installed.
    """
    try:
        __import__(package)
    except ImportError:
        raise ImportError(
            f"Package '{package}' is required but not installed. "
            f"Please install it using pip: pip install {package}"
        )


# Valid similarity measures and transformations as listed in the module docstring
VALID_METHODS = [
    "Jaccard",
    "Kulczynski2",
    "Russel",
    "Faith",
    "Tanimoto",
    "Dice",
    "Mozley",
    "Ochiai",
    "Simpson",
    "cosine",
    "angular",
    "correlation",
    "Tschuprow",
    "Cramer",
    "Gower",
    "Euclidean",
    "Manhattan",
    "supremum",
    "Canberra",
    "Chord",
    "Geodesic",
    "Bray",
    "Soergel",
    "Podani",
    "Whittaker",
    "eJaccard",
    "eDice",
    "Bhjattacharyya",
    "divergence",
    "Hellinger",
    "edit_sim_utf8",
    "edit_sim",
    "Levenshtein",
    "sim_NCD",
    "const",
    "sim_dtw",
]

VALID_TRANSFORMATIONS = [
    "pitch",
    "int",
    "fuzzy_int",
    "parsons",
    "pc",
    "ioi_class",
    "duration_class",
    "int_X_ioi_class",
    "implicit_harmonies",
]


def validate_method(method: str):
    """Validate that the similarity method is supported.

    Parameters
    ----------
    method : str
        Name of the similarity method to validate.

    Raises
    ------
    ValueError
        If the method is not supported.
    """
    if method not in VALID_METHODS:
        raise ValueError(
            f"Invalid method '{method}'. Valid methods are: {', '.join(VALID_METHODS)}"
        )


def validate_transformation(transformation: str):
    """Validate that the transformation is supported.

    Parameters
    ----------
    transformation : str
        Name of the transformation to validate.

    Raises
    ------
    ValueError
        If the transformation is not supported.
    """
    if transformation not in VALID_TRANSFORMATIONS:
        raise ValueError(
            f"Invalid transformation '{transformation}'. Valid transformations are: {', '.join(VALID_TRANSFORMATIONS)}"
        )


def get_similarity(
    melody_1, melody_2, method: str, transformation: str
) -> float:
    """Calculate similarity between two Score objects using the specified method.

    Parameters
    ----------
    melody_1 : Score
        First Score object containing a monophonic melody
    melody_2 : Score
        Second Score object containing a monophonic melody
    method : str
        Name of the similarity method to use from the list in the module docstring.
    transformation : str
        Name of the transformation to use from the list in the module docstring.

    Returns
    -------
    float
        Similarity value between the two melodies

    Examples
    --------
    >>> from amads.core.basics import Score
    >>> # Create two simple melodies using from_melody
    >>> melody_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    >>> melody_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)
    >>> # Calculate similarity using Jaccard method
    >>> similarity = get_similarity(melody_1, melody_2, 'Jaccard', 'pitch')
    """
    # Validate inputs
    validate_method(method)
    validate_transformation(transformation)

    # Convert Score objects to arrays
    pitches1, starts1, ends1 = score_to_arrays(melody_1)
    pitches2, starts2, ends2 = score_to_arrays(melody_2)

    # Pass lists directly to _get_similarity
    return _get_similarity(
        pitches1,
        starts1,
        ends1,
        pitches2,
        starts2,
        ends2,
        method,
        transformation,
    )


def _get_similarity(
    melody1_pitches: List[float],
    melody1_starts: List[float],
    melody1_ends: List[float],
    melody2_pitches: List[float],
    melody2_starts: List[float],
    melody2_ends: List[float],
    method: str,
    transformation: str,
) -> float:
    """Calculate similarity between two melodies using the specified method.
    Parameters
    ----------
    melody1_pitches : List[float]
        Pitch values for the first melody
    melody1_starts : List[float]
        Start times for the first melody
    melody1_ends : List[float]
        End times for the first melody
    melody2_pitches : List[float]
        Pitch values for the second melody
    melody2_starts : List[float]
        Start times for the second melody
    melody2_ends : List[float]
        End times for the second melody
    method : str
        Name of the similarity method to use
    transformation : str
        Name of the transformation to use
    Returns
    -------
    float
        Similarity value between the two melodies
    """
    # Validate inputs
    validate_method(method)
    validate_transformation(transformation)

    # Convert arrays to comma-separated strings (works with both lists and numpy arrays)
    pitches1_str = ",".join(map(str, melody1_pitches))
    starts1_str = ",".join(map(str, melody1_starts))
    ends1_str = ",".join(map(str, melody1_ends))
    pitches2_str = ",".join(map(str, melody2_pitches))
    starts2_str = ",".join(map(str, melody2_starts))
    ends2_str = ",".join(map(str, melody2_ends))

    # Create R script for similarity calculation
    r_script = f"""
    suppressMessages(suppressWarnings({{
        library(melsim)
        # Create melody objects
        melody1 <- melody_factory$new(mel_data = tibble::tibble(
            onset = c({starts1_str}),
            pitch = c({pitches1_str}),
            duration = c({ends1_str}) - c({starts1_str})
        ))
        melody2 <- melody_factory$new(mel_data = tibble::tibble(
            onset = c({starts2_str}),
            pitch = c({pitches2_str}),
            duration = c({ends2_str}) - c({starts2_str})
        ))
        # Create similarity measure
        sim_measure <- sim_measure_factory$new(
            name = "{method}",
            full_name = "{method}",
            transformation = "{transformation}",
            parameters = list(),
            sim_measure = "{method}"
        )
        # Calculate similarity
        result <- melody1$similarity(melody2, sim_measure)
        cat(jsonlite::toJSON(result$sim))
    }}))
    """

    # Run R script
    try:
        output_str = run_script_in_r(r_script)
        # Extract JSON from output (may contain warnings before the JSON)
        # Find the JSON part - look for the last line or the part after newline
        lines = output_str.split("\n")
        json_str = lines[-1]  # JSON should be on the last line

        # If that doesn't work, try to find JSON array/object markers
        if not (
            json_str.startswith("[")
            or json_str.startswith("{")
            or json_str.startswith('"')
        ):
            for line in reversed(lines):
                if (
                    line.startswith("[")
                    or line.startswith("{")
                    or line.startswith('"')
                ):
                    json_str = line
                    break

        output = json.loads(json_str)

        # Handle both single values and lists
        if isinstance(output, list):
            value = output[0]  # Get first value if it's a list
        else:
            value = output

        # Handle NA values from R
        if value == "NA" or value is None:
            return float("nan")  # Return NaN for invalid combinations

        return float(value)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error calculating similarity: {e.stderr}")


def _convert_strings_to_tuples(d: Dict) -> Dict:
    """Convert string keys back to tuples where needed."""
    result = {}
    for k, v in d.items():
        if isinstance(v, dict):
            result[k] = _convert_strings_to_tuples(v)
        else:
            result[k] = v
    return result


def score_to_arrays(score) -> Tuple[List[float], List[float], List[float]]:
    """Extract melody attributes from a Score object.

    Parameters
    ----------
    score : Score
        Score object containing a monophonic melody

    Returns
    -------
    Tuple[List[int], List[float], List[float]]
        Tuple of (pitches, start_times, end_times)
    """
    assert score.ismonophonic(), "Score must be monophonic"

    notes = score.get_sorted_notes()

    # Extract onset, pitch, duration for each note
    pitches = [note.pitch.key_num for note in notes]
    starts = [note.onset for note in notes]
    ends = [note.onset + note.duration for note in notes]

    return pitches, starts, ends


def _batch_compute_similarities(args_list: List[Tuple]) -> List[float]:
    """Compute similarities for a batch of melody pairs.

    Parameters
    ----------
    args_list : List[Tuple]
        List of argument tuples for _compute_similarity

    Returns
    -------
    List[float]
        List of similarity values
    """
    # Create R script for batch similarity calculation with improved efficiency
    r_script = """
    suppressMessages(suppressWarnings({
        library(melsim)
        library(jsonlite)
        library(purrr)
    # Function to create melody object
    create_melody <- function(pitches, starts, ends) {
        melody_factory$new(mel_data = tibble::tibble(
            onset = as.numeric(strsplit(starts, ",")[[1]]),
            pitch = as.numeric(strsplit(pitches, ",")[[1]]),
            duration = as.numeric(strsplit(ends, ",")[[1]]) - as.numeric(strsplit(starts, ",")[[1]])
        ))
    }
    # Function to calculate similarity
    calc_similarity <- function(melody1, melody2, method, transformation) {
        sim_measure <- sim_measure_factory$new(
            name = method,
            full_name = method,
            transformation = transformation,
            parameters = list(),
            sim_measure = method
        )
        result <- melody1$similarity(melody2, sim_measure)
        result$sim
    }
    # Process command line arguments
    args <- commandArgs(trailingOnly = TRUE)
    n_args <- length(args)
    n_comparisons <- n_args / 8  # Each comparison has 8 arguments
    # Pre-allocate results vector
    results <- numeric(n_comparisons)
    # Create a cache for melody objects
    melody_cache <- new.env()
    # Process in chunks for better memory management
    chunk_size <- 1000
    n_chunks <- ceiling(n_comparisons / chunk_size)
    for (chunk in seq_len(n_chunks)) {
        start_idx <- (chunk - 1) * chunk_size + 1
        end_idx <- min(chunk * chunk_size, n_comparisons)
        # Process chunk
        for (i in start_idx:end_idx) {
            idx <- (i-1) * 8 + 1
            # Get or create melody1
            melody1_key <- paste(args[idx], args[idx+1], args[idx+2], sep="|")
            if (!exists(melody1_key, envir=melody_cache)) {
                melody_cache[[melody1_key]] <- create_melody(args[idx], args[idx+1], args[idx+2])
            }
            melody1 <- melody_cache[[melody1_key]]
            # Get or create melody2
            melody2_key <- paste(args[idx+3], args[idx+4], args[idx+5], sep="|")
            if (!exists(melody2_key, envir=melody_cache)) {
                melody_cache[[melody2_key]] <- create_melody(args[idx+3], args[idx+4], args[idx+5])
            }
            melody2 <- melody_cache[[melody2_key]]
            method <- args[idx+6]
            transformation <- args[idx+7]
            results[i] <- calc_similarity(melody1, melody2, method, transformation)
        }
        # Force garbage collection after each chunk
        gc()
    }
    cat(toJSON(results))
    }))
    """

    # Prepare all arguments
    all_args = []
    for melody1_data, melody2_data, method, transformation in args_list:
        # Convert lists to comma-separated strings
        pitches1_str = ",".join(map(str, melody1_data[0]))
        starts1_str = ",".join(map(str, melody1_data[1]))
        ends1_str = ",".join(map(str, melody1_data[2]))
        pitches2_str = ",".join(map(str, melody2_data[0]))
        starts2_str = ",".join(map(str, melody2_data[1]))
        ends2_str = ",".join(map(str, melody2_data[2]))

        all_args.extend(
            [
                pitches1_str,
                starts1_str,
                ends1_str,
                pitches2_str,
                starts2_str,
                ends2_str,
                method,
                transformation,
            ]
        )

    # Run R script with all arguments
    try:
        output_str = run_script_in_r(r_script)
        # Extract JSON from output (may contain warnings before the JSON)
        # Find the JSON part - look for the last line or the part after newline
        lines = output_str.split("\n")
        json_str = lines[-1]  # JSON should be on the last line

        # If that doesn't work, try to find JSON array/object markers
        if not (
            json_str.startswith("[")
            or json_str.startswith("{")
            or json_str.startswith('"')
        ):
            for line in reversed(lines):
                if (
                    line.startswith("[")
                    or line.startswith("{")
                    or line.startswith('"')
                ):
                    json_str = line
                    break

        parsed_result = json.loads(json_str)
        # Handle NA values from R
        return [
            float("nan") if x == "NA" or x is None else float(x)
            for x in parsed_result
        ]
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error calculating similarities: {e.stderr}")


def get_similarities(
    scores: Dict[str, object],
    method: Union[str, List[str]] = "Jaccard",
    transformation: Union[str, List[str]] = "pitch",
    output_file: Union[str, Path, None] = None,
    n_cores: Optional[int] = None,
    batch_size: int = 1000,
) -> Union[
    Dict[str, Dict[str, float]],
    Dict[Tuple[str, str], Dict[str, Dict[str, float]]],
]:
    """Calculate pairwise similarities between multiple Score objects.

    You can provide a single method and transformation, or a list of methods and transformations.
    The function will return similarity matrices as nested dictionaries.

    Parameters
    ----------
    scores : Dict[str, Score]
        Dictionary mapping score names to Score objects
    method : Union[str, List[str]], default="Jaccard"
        Name of the similarity method(s) to use. Can be a single method or a list of methods.
    transformation : Union[str, List[str]], default="pitch"
        Name of the transformation(s) to use. Can be a single transformation or a list of transformations.
    output_file : Union[str, Path], optional
        If provided, save results to this file. If no extension is provided, .json will be added.
    n_cores : int, optional
        Number of CPU cores to use for parallel processing. Defaults to all available cores.
    batch_size : int, default=1000
        Number of comparisons to process in each batch

    Returns
    -------
    Union[Dict[str, Dict[str, float]], Dict[Tuple[str, str], Dict[str, Dict[str, float]]]]
        If single method and transformation: nested dictionary similarity
        matrix {row_name: {col_name: similarity}} where row_name and col_name
        are score names. If multiple methods/transformations: dictionary mapping
        (method, transformation) tuples to similarity matrices
    """
    # Convert single method/transformation to lists
    methods = [method] if isinstance(method, str) else method
    transformations = (
        [transformation] if isinstance(transformation, str) else transformation
    )

    # Validate all methods and transformations
    for m in methods:
        validate_method(m)
    for t in transformations:
        validate_transformation(t)

    if len(scores) < 2:
        raise ValueError("Need at least 2 Score objects for comparison")

    # Extract melody data from all scores (avoid multiprocessing due to Score object pickling issues)
    print("Extracting melody data...")
    melody_data = {}
    for name, score in tqdm(scores.items(), desc="Processing Score objects"):
        try:
            melody_data[name] = score_to_arrays(score)
        except Exception as e:
            print(
                f"Warning: Could not extract melody data for {name}: {str(e)}"
            )

    if len(melody_data) < 2:
        raise ValueError("Need at least 2 valid Score objects for comparison")

    # Prepare arguments for parallel processing
    print("Computing similarities...")
    args = []
    score_pairs = []

    # Pre-compute all combinations for better performance
    combinations_list = list(combinations(melody_data.items(), 2))
    for (name1, data1), (name2, data2) in combinations_list:
        for m in methods:
            for t in transformations:
                args.append((data1, data2, m, t))
                score_pairs.append((name1, name2, m, t))

    # Process in larger batches for better performance
    similarities_list = []
    for i in tqdm(range(0, len(args), batch_size), desc="Processing batches"):
        batch = args[i : i + batch_size]
        similarities_list.extend(_batch_compute_similarities(batch))

    # Create dictionary of results
    similarities = dict(zip(score_pairs, similarities_list))

    # Convert to matrix format using native Python types
    score_names = list(scores.keys())

    # Create similarity matrices as nested dictionaries
    matrices = {}

    for m in methods:
        for t in transformations:
            # Initialize matrix as nested dictionary with 1s on diagonal
            sim_matrix = {}
            for name1 in score_names:
                sim_matrix[name1] = {}
                for name2 in score_names:
                    if name1 == name2:
                        sim_matrix[name1][name2] = 1.0
                    else:
                        sim_matrix[name1][name2] = 0.0

            # Fill matrix with pairwise similarities
            # Since combinations() only gives us each pair once, set both directions
            for (
                name1,
                name2,
                method_key,
                transformation_key,
            ), similarity in similarities.items():
                if method_key == m and transformation_key == t:
                    # Handle NaN values consistently
                    if (
                        similarity == "NA"
                        or similarity is None
                        or (
                            isinstance(similarity, float)
                            and math.isnan(similarity)
                        )
                    ):
                        sim_value = float("nan")
                    else:
                        sim_value = float(similarity)

                    # Set both directions to ensure perfect symmetry
                    sim_matrix[name1][name2] = sim_value
                    sim_matrix[name2][name1] = sim_value

            matrices[(m, t)] = sim_matrix

    # Save to file if output file specified
    if output_file:
        print("Saving results...")

        # Ensure output file has .json extension
        output_file = Path(output_file)
        if not output_file.suffix:
            output_file = output_file.with_suffix(".json")

        # Save matrices to JSON
        output_data = {}
        for (m, t), matrix in matrices.items():
            output_data[f"{m}_{t}"] = matrix

        import json

        with open(output_file, "w") as f:
            json.dump(output_data, f, indent=2)
        print(f"Results saved to {output_file}")

    # Return format depends on number of method/transformation combinations
    if len(methods) == 1 and len(transformations) == 1:
        # Single method and transformation: return just the matrix
        return matrices[(methods[0], transformations[0])]
    else:
        # Multiple methods/transformations: return dictionary of matrices
        return matrices



"""
Tests for meter representation.

Tests functionality for regrouping and quantizing musical durations.
"""

import pytest

from amads.time.meter import examples
from amads.time.meter.representations import (
    PulseLengths,
    StartTimeHierarchy,
    TimeSignature,
)

metres = []
for x in (
    (
        "4/4",
        [1, 2],
        [4, 2, 1],
        [[0.0, 4.0], [0.0, 2.0, 4.0], [0.0, 1.0, 2.0, 3.0, 4.0]],
    ),
    (
        "4/4",
        [3],
        [4, 0.5],
        [[0.0, 4.0], [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0]],
    ),
    (
        "6/8",
        [1, 2],
        [3, 1.5, 0.5],
        [[0.0, 3.0], [0.0, 1.5, 3.0], [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0]],
    ),
):
    new_dict = {"ts": x[0], "levels": x[1], "pulses": x[2], "starts": x[3]}
    metres.append(new_dict)


# def test_pulse_lengths_to_start_list(test_metres):
#     for tc in test_metres:
#         start_hierarchy = PulseLengths(pulse_lengths=tc["pulses"]).to_start_hierarchy()
#         assert start_hierarchy == tc["starts"]


def test_start_hierarchy_from_time_signature():
    """Test start_hierarchy_from_time_signature by running through test cases."""
    for example in examples.start_hierarchy_examples:
        ts = TimeSignature(as_string=example)
        ts.fill_2s_3s()
        sh = ts.to_start_hierarchy()
        sh = StartTimeHierarchy(start_hierarchy=sh)
        sh.add_faster_levels(minimum_beat_type=32)
        assert sh.start_hierarchy == examples.start_hierarchy_examples[example]


"""Test various cases that should raise errors and similar."""


def test_invalid_denominator():
    with pytest.raises(ValueError):
        TimeSignature(as_string="2/6")


def test_pulse_beyond_cycle_length():
    with pytest.raises(ValueError):
        PulseLengths([4, 2, 1], cycle_length=2)


def test_name_format():
    """
    One case in the correct format, and one that raises.
    """
    StartTimeHierarchy(
        [[0.0, 0.1]], names={0.0: "ta", 1.0: "ka", 2.0: "di", 3.0: "mi"}
    )

    with pytest.raises(AssertionError):
        StartTimeHierarchy([[0.0, 0.1]], names="Aditya, Bella, Carlos")

    with pytest.raises(AssertionError):
        StartTimeHierarchy(
            [[0.0, 0.1]], names={0.0: ["Aditya", "Bella", "Carlos"]}
        )



"""Test suite for amads.time.swing.py"""

import numpy as np
import pytest

from amads.time.swing import (
    _validate_bur_inputs,
    beat_upbeat_ratio,
    mean_bur,
    std_bur,
)

TEST_BEATS = [
    [0.0, 1.0, 2.0],
    [0.0, 1.0, 2.0, 3.0, 4.0],
    [0.0, 1.3, 2.2, 3.3, 5.5, 6.1],
]
TEST_UPBEATS = [
    [0.6, 1.4],  # example 1, straightforward
    [0.5, 1.2, 2.6, 3.5, 4.9],  # example 2, slightly harder
    [
        0.5,
        1.2,
        2.7,
        3.5,
        3.6,
    ],  # example 3: b1 + b4 skipped as two upbeats, b2 + b5 skipped as have no upbeat
]
TEST_BURS = [
    [1.5, 2 / 3],
    [1.0, 0.25, 1.5, 1.0],
    [None, None, 0.5 / 0.6, None, None],
]


@pytest.mark.parametrize(
    "beats,upbeats,expecteds", zip(TEST_BEATS, TEST_UPBEATS, TEST_BURS)
)
def test_bur(beats, upbeats, expecteds):
    burs = beat_upbeat_ratio(beats, upbeats)
    assert pytest.approx(burs) == expecteds
    assert len(burs) == len(beats) - 1


def test_validate_bur_inputs():
    # Test 1: Invalid input - Less than two beats
    beats = np.array([1.0])
    upbeats = np.array([1.5])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)

    # Test 2: Invalid input - No upbeats
    beats = np.array([1.0, 2.0, 3.0])
    upbeats = np.array([])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)

    # Test 3: Invalid input - Multi-dimensional beats
    beats = np.array([[1.0, 2.0], [3.0, 4.0]])
    upbeats = np.array([1.5])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)

    # Test 4: Invalid input - Multi-dimensional upbeats
    beats = np.array([1.0, 2.0, 3.0])
    upbeats = np.array([[1.5]])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)

    # Test 5: Invalid input - NaN in beats
    beats = np.array([1.0, np.nan, 3.0])
    upbeats = np.array([1.5])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)

    # Test 6: Invalid input - NaN in upbeats
    beats = np.array([1.0, 2.0, 3.0])
    upbeats = np.array([np.nan])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)

    # Test 7: Invalid input - Intersection between beats and upbeats
    beats = np.array([1.0, 2.0, 3.0])
    upbeats = np.array([2.0])
    with pytest.raises(ValueError):
        _validate_bur_inputs(beats, upbeats)


def test_no_matches():
    beats = [
        0.0,
        1.0,
        2.0,
        3.0,
    ]
    upbeats = [5.0, 6.0, 7.0, 8.0]
    with pytest.raises(ValueError):
        _ = beat_upbeat_ratio(beats, upbeats)


def test_bounded_burs():
    beats = [
        0.0,
        1.0,
        2.0,
        3.0,
    ]
    upbeats = [0.5, 1.5, 2.99]
    expected = [1.0, 1.0, None]
    actual = beat_upbeat_ratio(beats, upbeats, bounded=True)
    assert pytest.approx(actual) == expected


def test_log2_burs():
    beats = [0.0, 1.0, 2.0, 3.0]
    upbeats = [1 / 3, 5 / 3, 2.5]
    expected = [-1.0, 1.0, 0.0]
    actual = beat_upbeat_ratio(beats, upbeats, log2=True)
    assert pytest.approx(actual) == expected


def test_agg_funcs():
    beats = [
        0.0,
        1.0,
        2.0,
        3.0,
    ]
    upbeats = [0.5, 1.75, 2.8]
    mean = 8 / 3  # burs = [1., 3., 4.]
    sd = 1.247219
    assert pytest.approx(mean_bur(beats, upbeats)) == mean
    assert pytest.approx(std_bur(beats, upbeats)) == sd



"""
Distribution of durations in a Score.
Provides the `ivdirdist1` function

Can emulate the `ivdurdist1` function in Midi Toolbox.

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=64
"""

from amads.core.basics import Score
from amads.core.distribution import Distribution
from amads.pitch.ivdist1 import interval_distribution_1


def interval_direction_distribution_1(
    score: Score,
    name: str = "Interval Direction Distribution",
    weighted: bool = True,
    miditoolbox_compatible: bool = True,
) -> Distribution:
    """
    Returns the proportion of upward intervals for each interval size

    Currently, intervals greater than an octave will be ignored.

    Parameters
    ----------
    score : Score
        The music Score object to analyze
    name : str
        A name for the resulting distribution (title in distribution plot)
    weighted : bool, optional
        If True, the interval distribution is weighted by note durations
        in seconds that are modified according to Parncutt's durational
        accent model (1994), by default True.
    miditoolbox_compatible : bool
        Invokes interval_distribution_1 using miditoolbox_compatible=True,
        which performs normalization slightly differently (see
        [interval_distribution_1]
        [amads.pitch.ivdist1.interval_distribution_1].
        Default is False, which simply skips division when the total
        count is zero (this also returns a zero matrix when the count
        is zero).

    Returns
    -------
    Distribution
        A 12-element distribution representing the proportion of
        upward intervals for each interval size. The components
        are spaced at semitone distances with the first component
        representing a minor second (not unison) and the last
        component the octave. If the score is empty, the function
        returns a list with all elements set to zero.
    """

    id = interval_distribution_1(score, name, weighted, miditoolbox_compatible)
    id = id.data  # we only need the data from the distribution
    idd = [0.0] * 12

    for i in range(12):
        # id[i + 13] is the upward interval
        # id[11 - i] is the downward interval
        if (id[i + 13] + id[11 - i]) != 0:
            idd[i] = id[i + 13] / (id[i + 13] + id[11 - i])
        else:
            idd[i] = 0

    x_categories = [str(i) for i in range(1, 13)]
    return Distribution(
        name,
        idd,
        "interval_direction",
        [12],
        x_categories,  # type: ignore
        "Interval Size",
        None,
        "Proportion",
    )



"""
Pitch interval analysis.

Implements the Midi Toolbox `ivdist1` function

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=63
"""

from typing import cast

from amads.core.basics import Note, Part, Score
from amads.core.distribution import Distribution
from amads.core.histogram import Histogram1D
from amads.pitch.pcdist1 import duraccent


def interval_distribution_1(
    score: Score,
    name: str = "Interval Distribution",
    weighted: bool = True,
    miditoolbox_compatible: bool = True,
) -> Distribution:
    """
    Returns the interval distribution of a musical score.

    Currently, intervals greater than an octave will be ignored.

    Parameters
    ----------
    score : Score
        The musical score to analyze
    name : str
        A name for the distribution and plot title.
    weighted : bool, optional
        If True, the interval distribution is weighted by note durations
        in seconds that are modified according to Parncutt's durational
        accent model (1994), by default True.
    miditoolbox_compatible : bool
        Matlab MIDI Toolbox avoids zero division by dividing counts
        by the total count plus (1e-12 times the number of bins).
        True enables this behavior. Default is False, which simply skips
        division when the total count is zero (this also returns a
        zero matrix when the count is zero).

    Returns
    -------
    Distribution
        A 25-bin distribution representing the probabilities of each pitch
        interval. The bins are spaced at semitone distances with the first
        bin representing the downward octave and the last bin representing
        the upward octave. If the score is empty, the function returns a
        list with all elements set to zero.

    Raises
    ------
    ValueError
        If the score is not monophonic (e.g. contains chords)
    """
    if not score.ismonophonic():
        raise ValueError("Error: Score must be monophonic")

    score = cast(Score, score.merge_tied_notes())
    if weighted:
        score.convert_to_seconds()

    bin_centers = [float(i - 12) for i in range(25)]  # 25 bins from -12 to +12
    bin_boundaries = [i - 12 - 0.5 for i in range(26)]  # boundaries
    x_categories = [str(c) for c in bin_centers]
    h = Histogram1D(bin_centers, bin_boundaries, "linear", True)

    for p in score.find_all(Part):
        part: Part = cast(Part, p)
        prev_pitch = None
        prev_dur = None
        for n in part.find_all(Note):
            note: Note = cast(Note, n)
            if prev_pitch is not None:
                iv = round(note.key_num - prev_pitch)
                if miditoolbox_compatible:
                    iv = (abs(iv) % 12) * ((iv > 0) - (iv < 0))
                # otherwise, diff may be ignored by h.add_point
                if weighted:
                    dur = duraccent(note)
                    # prev_dur cannot be None here since prev_pitch is not None
                    h.add_point(iv, prev_dur + dur)  # type: ignore
                    prev_dur = dur
                else:
                    h.add_point(iv, 1.0)
            prev_pitch = note.key_num
            if weighted and prev_dur is None:
                prev_dur = duraccent(note)

    if miditoolbox_compatible:  # miditoolbox "normalization"
        total = sum(h.bins) + len(h.bins) * 1e-12
        h.bins = [b / total for b in h.bins]
    else:  # normalize normally
        h.normalize()

    return Distribution(
        name,
        h.bins,
        "interval",
        [len(h.bins)],
        x_categories,  # type: ignore
        "Interval (semitones)",
        None,
        "Proportion",
    )



import warnings

import pytest

from amads.algorithms.mtype_tokenizer import FantasticTokenizer, MType
from amads.algorithms.ngrams import NGramCounter
from amads.core.basics import Score
from amads.melody.segment import fantastic_segmenter


def test_mtype_tokenizer():
    melody_1 = Score.from_melody(
        pitches=[62, 64, 65, 67, 64, 60, 62],  # D4, E4, F4, G4, E4, C4, D4
        durations=[
            0.5,
            0.5,
            0.5,
            0.5,
            1.0,
            0.5,
            0.5,
        ],  # quarter, eighth, quarter, eighth, eighth notes
    )

    # Initialize tokenizer and counter
    tokenizer = FantasticTokenizer()
    ngrams = NGramCounter()

    # Tokenize melody and count n-grams
    tokens = tokenizer.tokenize(melody_1)

    # Count bigrams
    ngrams.count_ngrams(tokens, n=2)
    for ngram in ngrams.get_counts():
        assert len(ngram) == 2

    # Reset counter, count trigrams
    ngrams.reset()
    ngrams.count_ngrams(tokens, n=3)
    for ngram in ngrams.get_counts():
        assert len(ngram) == 3

    # Test that n-grams cannot be counted if method is larger than sequence length
    with pytest.raises(ValueError):
        ngrams.count_ngrams(tokens, n=10)

    # Test that n-grams cannot be counted if method n = 0
    with pytest.raises(ValueError):
        ngrams.count_ngrams(tokens, n=0)

    # Test that n=None returns all n-grams
    ngrams.reset()
    ngrams.count_ngrams(tokens, n=None)
    all_counts = ngrams.get_counts()
    # Test each n-gram length by comparing to individual n-gram counts
    max_length = max(len(ngram) for ngram in all_counts)
    for n in range(1, max_length + 1):
        ngrams.reset()
        ngrams.count_ngrams(tokens, n=n)
        n_gram_counts = ngrams.get_counts()
        # Check that all n-grams of length n in n_gram_counts are also in all_counts
        for ngram in n_gram_counts:
            assert ngram in all_counts
            assert n_gram_counts[ngram] == all_counts[ngram]

    # Test with a second melody using a different phrase gap
    tokenizer = FantasticTokenizer()
    ngrams = NGramCounter()

    # Create a test melody which will be segmented into 2 phrases
    melody_2 = Score.from_melody(
        # Twinkle Twinkle Little Star
        pitches=[60, 60, 67, 67, 69, 69, 67, 65, 65, 64, 64, 62, 62, 60],
        durations=[
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            1.0,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
        ],
    )

    # Test that the melody is segmented into phrases
    segments = fantastic_segmenter(melody_2, phrase_gap=0.75, units="quarters")

    # Test counting all n-grams across segments
    ngrams.reset()
    for segment in segments:
        segment_tokens = tokenizer.tokenize(segment)
        ngrams.count_ngrams(segment_tokens, n=None)

    # Get all n-gram counts
    all_counts = ngrams.get_counts()

    # Test that we can't count n-grams longer than sequence
    max_length = max(len(ngram) for ngram in all_counts)

    # As the phrase length is 7 notes, the maximum length n-gram is 6
    assert max_length == 6
    with pytest.raises(ValueError):
        ngrams.count_ngrams(segment_tokens, n=max_length + 1)


def test_mtype_encodings():
    possible_interval_classes = FantasticTokenizer.interval_classes

    assert "d3" in possible_interval_classes
    assert "u5" in possible_interval_classes
    assert None in possible_interval_classes

    possible_ioi_ratio_classes = FantasticTokenizer.ioi_ratio_classes

    assert "q" in possible_ioi_ratio_classes
    assert "e" in possible_ioi_ratio_classes
    assert "l" in possible_ioi_ratio_classes
    assert None in possible_ioi_ratio_classes

    mtypes = []
    for interval_class in possible_interval_classes:
        for ioi_ratio_class in possible_ioi_ratio_classes:
            mtype = MType(interval_class, ioi_ratio_class)
            mtypes.append(mtype)

            assert isinstance(mtype.integer, int)
            assert mtype.integer >= 0

    assert len(mtypes) == len(set(mtypes))

    integers = [mtype.integer for mtype in mtypes]
    assert len(integers) == len(set(integers))

    for integer in integers:
        assert 0 <= integer <= len(integers) - 1


def test_ngram_counts():

    simple_list = [0, 1, 1, 0, 1]
    complex_list = [0, 1, 2, 3, 4, 5]
    simple_ngrams = NGramCounter()
    complex_ngrams = NGramCounter()

    # Features calculated from the ngrams are not available until ngrams are counted
    with pytest.raises(ValueError):
        _ = simple_ngrams.yules_k

    # Count the ngrams
    simple_ngrams.count_ngrams(simple_list, n=2)
    assert simple_ngrams.get_counts() == {
        ("0", "1"): 2,
        ("1", "1"): 1,
        ("1", "0"): 1,
    }

    # Now the features are available
    assert simple_ngrams.yules_k is not None

    complex_ngrams.count_ngrams(complex_list, n=2)

    # We expect that Yule's K is lower for the complex list, as there is more variety in the n-grams
    # Yule's K takes a higher value for more repetitive sequences
    assert complex_ngrams.yules_k < simple_ngrams.yules_k

    # Simpson's D should also be higher for more repetitive sequences
    assert complex_ngrams.simpsons_d < simple_ngrams.simpsons_d

    # Sichel's S should also be higher for more the simple list,
    # as complex_list contains no n-grams that occur exactly twice
    assert complex_ngrams.sichels_s < simple_ngrams.sichels_s

    # Honore's H is incalculable for complex_list,
    # as the number of hapax legomena is equal to the number of total types
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        _ = complex_ngrams.honores_h

    # Normalized entropy should be lower for more repetitive sequences
    assert complex_ngrams.mean_entropy > simple_ngrams.mean_entropy

    # Mean productivity should be higher for more diverse sequences
    assert complex_ngrams.mean_productivity > simple_ngrams.mean_productivity


if __name__ == "__main__":
    test_mtype_tokenizer()
    test_mtype_encodings()
    test_ngram_counts()



"""
Prepare a grid
based on the tatum (smallest value needed to fully express all event positions).
This will be a monotonically increasing sequence of integers
where each integer expresses the event position relative in multiples of the tatum.

This is needed for algorithms including the
Inner Metrical Analysis (IMA) which takes this list of integers indices
and deduced local and spectral meters.

<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"

from fractions import Fraction

from amads.io.readscore import read_score, set_reader_warning_level
from amads.time.meter.grid import get_tatum_from_priorities


def score_to_offsets(path_to_score: str, to_indices: bool = True) -> list:
    """
    Import a score and convert it to the sorted list of unique
    starting timepoints as measured in quarters since the start of the score,
    and (optionally) convert those starts to indices on a tatum grid.

    Note: score parsing warnings are supressed.
    If you need to test the validity of scores, handle that separately.

    Parameters
    ----------
    path_to_score
        A string for the file path or URL.
    to_indices
        If True, convert the starts to indices on a tatum grid.

    Examples
    --------
    Two examples from "Species" Counterpoint.
    The first is straightforwardly in regular whole notes moving together,
    so the gaps are 4.0 apart (in "quarter notes") and the tatum is 4.

    >>> score_path = "https://github.com/MarkGotham/species/raw/refs/heads/main/1x1/005.mxl"
    >>> score_to_offsets(score_path, to_indices=False)
    [0.0, 4.0, 8.0, 12.0, 16.0, 20.0, 24.0, 28.0, 32.0, 36.0, 40.0]

    >>> score_to_offsets(score_path)
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    The second example is from later in book 1.
    This is the first example of "florid" (5th species) counterpoint.
    There is one pair of eighth notes in this example (offsets 29.0, 29.5, 30.0) so the tatum is 0.5.

    >>> url_5th_species = "https://github.com/MarkGotham/species/raw/refs/heads/main/1x1/082.mxl"
    >>> starts = score_to_offsets(url_5th_species, to_indices=False)

    This is how it starts:
    >>> starts[:5]
    [0.0, 2.0, 4.0, 5.0, 6.0]

    And this is the part with the eight note pair:
    >>> starts[22:]
    [28.0, 29.0, 29.5, 30.0, 32.0, 33.0, 34.0, 36.0, 38.0, 40.0]

    >>> indices = starts_to_indices(starts)
    >>> indices[:5]
    [0, 4, 8, 10, 12]

    >>> indices = score_to_offsets(url_5th_species, to_indices=True)
    >>> indices[:5]
    [0, 4, 8, 10, 12]

    >>> indices[22:]
    [56, 58, 59, 60, 64, 66, 68, 72, 76, 80]

    """
    set_reader_warning_level("none")
    score = read_score(path_to_score, show=False)
    notes = score.get_sorted_notes()
    timepoints = sorted(set(n.onset for n in notes))
    if to_indices:
        return starts_to_indices(timepoints)
    else:
        return timepoints


def starts_to_indices(starts: list, tatum: Fraction = None) -> list:
    """
    Given a list of start times,
    convert to a list of indices on the tatum grid.

    If a tatum value is provided, use that;
    otherwise, deduce the tatum using gcd methods.

    This is the input format for the IMA algorithm, among others.

    Parameters
    ----------
    starts
        A list of numeric start times.
    tatum
        The tatum duration to use as the grid unit and values are rounded to it.
        If None, it is
        deduced automatically via `get_tatum_from_priorities`.

    Examples
    --------
    >>> starts_to_indices([0, 1/2, 2/3, 2.5])
    [0, 3, 4, 15]

    >>> starts_to_indices([0, 1/2, 2/3, 2.5], tatum=Fraction(1, 6))
    [0, 3, 4, 15]

    >>> starts_to_indices([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 6.5, 6.6640625, 6.83203125, 7.0])
    [0, 6, 12, 18, 24, 30, 36, 39, 40, 41, 42]

    >>> starts_to_indices([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 6.5, 6.6640625, 6.83203125, 7.0], tatum=Fraction(1, 6))
    [0, 6, 12, 18, 24, 30, 36, 39, 40, 41, 42]

    Also accepts tatum values greater than 1:

    >>> starts_to_indices([3, 6, 9], tatum=3)
    [1, 2, 3]

    >>> starts_to_indices([3, 6, 9])
    [1, 2, 3]

    >>> starts_to_indices([0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40])
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    >>> starts_to_indices([28.0, 29.0, 29.5, 30.0, 32.0, 33.0, 34.0, 36.0, 38.0, 40.0])
    [56, 58, 59, 60, 64, 66, 68, 72, 76, 80]

    """
    if not starts:
        raise ValueError("starts must not be empty")

    if tatum is None:
        tatum = get_tatum_from_priorities(starts)

    return [round(x / tatum) for x in starts]


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Provides the `ivsizedist1` function
"""

from amads.core.basics import Score
from amads.core.distribution import Distribution
from amads.pitch.ivdist1 import interval_distribution_1


def interval_size_distribution_1(
    score: Score,
    name: str = "Interval Size Distribution",
    weighted: bool = True,
    miditoolbox_compatible: bool = True,
) -> Distribution:
    """
    Returns the interval size distribution of a musical score.

    Intervals greater than one octave are ignored.

    Parameters
    ----------
    score : Score
        The musical score to analyze
    name : str
        A name for the distribution and plot title.
    weighted : bool, optional
        If True, the interval distribution is weighted by note durations
        in seconds that are modified according to Parncutt's durational
        accent model (1994), by default True.
    miditoolbox_compatible : bool
        Invokes interval_distribution_1 using miditoolbox_compatible=True,
        which performs normalization slightly differently (see
        [interval_distribution_1]
        [amads.pitch.ivdist1.interval_distribution_1].
        Default is False, which simply skips division when the total
        count is zero (this also returns a zero matrix when the count
        is zero).

    Returns
    -------
    Distribution
        A 13-element distribution representing proportions of interval sizes.
        The first element corresponds to unison intervals, and the last
        element corresponds to octave intervals. If the score is empty,
        the function returns a Distribution with all elements set to zero.
    """
    id = interval_distribution_1(score, name, weighted, miditoolbox_compatible)
    id = id.data  # we only need the data from the distribution
    isd = [0.0] * 13

    isd[0] = id[12]
    for i in range(1, 13):
        isd[i] = id[i + 12] + id[12 - i]  # merge upward and downward bins
    # note that isd is normalized because it sums to the same value as isd
    x_categories = [str(i) for i in range(13)]
    return Distribution(
        name,
        isd,
        "interval_size",
        [12],
        x_categories,  # type: ignore
        "Interval Size",
        None,
        "Proportion",
    )



"""
Calculate the relative entropy of a distribution. (MTB)

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=60
"""

from typing import List

import numpy as np


def entropy(
    d: List[float], in_bits=False, miditoolbox_compatible: bool = False
) -> float:
    """
    Calculate the relative entropy of a distribution.

    This function computes the relative entropy (also known as normalized
    entropy) of a given distribution `d`. The result is a value between 0
    and 1, where 0 indicates no uncertainty (all probability mass is
    concentrated in one outcome) and 1 indicates maximum uncertainty
    (uniform distribution).

    If `in_bits` is set to True, the result will be in bits
    (base 2 logarithm) instead of relative entropy. The `in_bits`
    result gives the expected number of bits needed to encode an outcome
    drawn from the distribution.

    The distribution `d` will be normalized if it does not already sum
    to 1. If `d` is all zeros, the function will return 1.0 to avoid
    division by zero. If `miditoolbox_compatible` is set to True, division
    by zero is avoided by adding a small constant to the distribution,
    which will slightly alter the result. (This is to maintain compatibility
    with the original MIDI Toolbox implementation.)

    Parameters
    ----------
    d : List[float]
        The input distribution.

    in_bits : bool, optional
        If True, returns unnormalized entropy in bits (base 2 logarithm).
        Default is False: returns relative entropy (0 <= H <= 1).

    miditoolbox_compatible : bool, optional
        If True, uses the original MIDI Toolbox method of calculation.
        Default is False.

    Returns
    -------
    float
        The relative entropy (0 <= H <= 1) unless `in_bits` is True,
        in which case the unnormalized entropy in bits is returned.

    Notes
    -----
    Implementation based on the original MATLAB code from:
    https://github.com/miditoolbox/1.1/blob/master/miditoolbox/entropy.m

    Examples
    --------
    Entropy is maximized when all outcomes are equally likely:
    >>> entropy([0.5, 0.5])
    1.0

    Entropy is minimized when one outcome is certain:
    >>> entropy([0.0, 1.0])
    0.0
    """
    darray = np.asarray(d).flatten()  # Convert to a 1D numpy array
    sum = np.sum(darray)
    if miditoolbox_compatible:
        sum += 1e-12  # Avoid division by zero
    elif sum == 0:
        return 1.0  # Avoid division by zero; return maximum entropy
    darray = darray / sum  # Normalize
    if miditoolbox_compatible:
        logd = np.log(darray + 1e-12)  # Avoid log(0)
    else:
        logd = np.where(
            darray > 0, np.log(darray), 0
        )  # Compute log(d), treating log(0) as 0
    h = -np.sum(darray * logd)
    if in_bits:
        h = h / np.log(2)  # Unnormalized entropy in bits
    else:
        h = h / np.log(len(darray))  # Normalize to relative entropy
    return float(h)



import math
import os
from unittest.mock import Mock

import pytest

from amads.core.basics import Score
from amads.melody.similarity.melsim import (
    _convert_strings_to_tuples,
    check_python_package_installed,
    check_r_packages_installed,
    get_similarities,
    get_similarity,
    install_dependencies,
    install_r_package,
    score_to_arrays,
    validate_method,
    validate_transformation,
)


@pytest.fixture(scope="session")
def installed_melsim_dependencies():
    on_ci = os.environ.get("CI") is not None
    install_missing = on_ci
    check_r_packages_installed(install_missing=install_missing)


def test_melsim_import(installed_melsim_dependencies):
    """Test that melsim can be imported."""
    from amads.melody.similarity.melsim import get_similarity

    assert callable(get_similarity)


def test_example_usage():
    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)
    similarity = get_similarity(mel_1, mel_2, "Jaccard", "pitch")
    assert similarity == 0.6


def test_transformation_usage():
    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[62, 64, 66, 67], durations=1.0)
    # Melody 2 is a transposition of Melody 1 by 2 semitones
    similarity = get_similarity(mel_1, mel_2, "Jaccard", "pitch")
    # As a result, the similarity should not be 1.0
    assert similarity != 1.0

    # However, the similarity between the intervals should be 1.0
    similarity = get_similarity(mel_1, mel_2, "Jaccard", "int")
    assert similarity == 1.0


def test_validate_method():
    """Test method validation."""
    # Valid methods should not raise errors
    validate_method("Jaccard")
    validate_method("cosine")
    validate_method("Euclidean")

    # Invalid methods should raise ValueError
    with pytest.raises(ValueError, match="Invalid method"):
        validate_method("InvalidMethod")

    with pytest.raises(ValueError, match="Invalid method"):
        validate_method("random_method")


def test_validate_transformation():
    """Test transformation validation."""
    # Valid transformations should not raise errors
    validate_transformation("pitch")
    validate_transformation("int")
    validate_transformation("parsons")

    # Invalid transformations should raise ValueError
    with pytest.raises(ValueError, match="Invalid transformation"):
        validate_transformation("InvalidTransformation")

    with pytest.raises(ValueError, match="Invalid transformation"):
        validate_transformation("random_transform")


def test_check_python_package_installed():
    """Test Python package checking."""
    # Valid package should not raise error
    check_python_package_installed("os")
    check_python_package_installed("sys")

    # Invalid package should raise ImportError
    with pytest.raises(
        ImportError, match="Package 'nonexistent_package' is required"
    ):
        check_python_package_installed("nonexistent_package")


def test_score_to_arrays():
    """Test score to arrays conversion."""
    mel = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    pitches, starts, ends = score_to_arrays(mel)

    assert len(pitches) == 4
    assert len(starts) == 4
    assert len(ends) == 4
    assert pitches == [60, 62, 64, 65]
    assert all(isinstance(p, int) for p in pitches)
    assert all(isinstance(s, float) for s in starts)
    assert all(isinstance(e, float) for e in ends)


def test_convert_strings_to_tuples():
    """Test string to tuple conversion utility."""
    # Test with regular dict
    input_dict = {"key1": "value1", "key2": "value2"}
    result = _convert_strings_to_tuples(input_dict)
    assert result == input_dict

    # Test with nested dict
    input_dict = {"outer": {"inner": "value"}}
    result = _convert_strings_to_tuples(input_dict)
    assert result == {"outer": {"inner": "value"}}

    # The function currently doesn't convert string keys to tuples
    # It just preserves the dictionary structure


def test_get_similarity_error_handling():
    """Test error handling in get_similarity."""
    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)

    # Test invalid method
    with pytest.raises(ValueError, match="Invalid method"):
        get_similarity(mel_1, mel_2, "InvalidMethod", "pitch")

    # Test invalid transformation
    with pytest.raises(ValueError, match="Invalid transformation"):
        get_similarity(mel_1, mel_2, "Jaccard", "InvalidTransformation")


def test_install_r_package_cran(monkeypatch):
    """Test R package installation for CRAN packages."""
    mock_run = Mock()
    monkeypatch.setattr("subprocess.run", mock_run)

    # Test CRAN package installation
    install_r_package("jsonlite")

    # Should have called subprocess.run with Rscript
    mock_run.assert_called_once()
    call_args = mock_run.call_args[0][0]
    assert call_args[0].endswith("Rscript")
    assert "jsonlite" in call_args[2]


def test_install_r_package_github(monkeypatch):
    """Test R package installation for GitHub packages."""
    mock_run = Mock()
    monkeypatch.setattr("subprocess.run", mock_run)

    # Test GitHub package installation
    install_r_package("melsim")

    # Should have called subprocess.run with Rscript
    mock_run.assert_called_once()
    call_args = mock_run.call_args[0][0]
    assert call_args[0].endswith("Rscript")
    assert "remotes::install_github" in call_args[2]


def test_install_r_package_invalid():
    """Test error handling for invalid R packages."""
    with pytest.raises(ValueError, match="Unknown package type"):
        install_r_package("nonexistent_package")


def test_install_dependencies_success(monkeypatch):
    """Test successful dependency installation."""
    # Mock successful package checks (no missing packages)
    mock_run = Mock(return_value=Mock(stdout="", returncode=0))
    monkeypatch.setattr("subprocess.run", mock_run)

    # Should not raise any errors
    install_dependencies()

    # Should have called subprocess.run multiple times for checking packages
    assert mock_run.call_count >= 2  # At least CRAN and GitHub checks


def test_install_dependencies_with_missing_packages(monkeypatch):
    """Test dependency installation with missing packages."""
    # Mock missing packages on first call, then success
    mock_run = Mock(
        side_effect=[
            Mock(stdout='"jsonlite","dplyr"', returncode=0),  # Missing CRAN
            Mock(stdout="", returncode=0),  # Install success
            Mock(stdout='"melsim"', returncode=0),  # Missing GitHub
            Mock(stdout="", returncode=0),  # Install success
        ]
    )
    monkeypatch.setattr("subprocess.run", mock_run)

    install_dependencies()

    # Should have called subprocess.run multiple times
    assert mock_run.call_count == 4


def test_get_similarities_output_formats():
    """Test different output formats for get_similarities."""
    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)
    mel_3 = Score.from_melody(pitches=[67, 69, 71, 72], durations=1.0)

    scores = {"melody1": mel_1, "melody2": mel_2, "melody3": mel_3}

    # Test multiple methods, single transformation
    results = get_similarities(
        scores, method=["Jaccard", "Dice"], transformation="pitch"
    )
    assert isinstance(results, dict)
    assert ("Jaccard", "pitch") in results
    assert ("Dice", "pitch") in results

    # Test single method, multiple transformations
    results = get_similarities(
        scores, method="Jaccard", transformation=["pitch", "int"]
    )
    assert isinstance(results, dict)
    assert ("Jaccard", "pitch") in results
    assert ("Jaccard", "int") in results

    # Test multiple methods and transformations
    results = get_similarities(
        scores, method=["Jaccard", "Dice"], transformation=["pitch", "int"]
    )
    assert isinstance(results, dict)
    assert ("Jaccard", "pitch") in results
    assert ("Jaccard", "int") in results
    assert ("Dice", "pitch") in results
    assert ("Dice", "int") in results


def test_get_similarities_batch_processing():
    """Test batch processing parameters."""
    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)

    scores = {"melody1": mel_1, "melody2": mel_2}

    # Test with different batch sizes
    results1 = get_similarities(
        scores, method="Jaccard", transformation="pitch", batch_size=1
    )
    results2 = get_similarities(
        scores, method="Jaccard", transformation="pitch", batch_size=10
    )

    # Results should be the same regardless of batch size
    assert results1["melody1"]["melody2"] == results2["melody1"]["melody2"]

    # Test with different n_cores
    results3 = get_similarities(
        scores, method="Jaccard", transformation="pitch", n_cores=1
    )
    assert results3["melody1"]["melody2"] == results1["melody1"]["melody2"]


def test_melsim_measures_transformations():

    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)

    supported_measures = [
        "Jaccard",
        "Kulczynski2",
        "Russel",
        "Faith",
        "Tanimoto",
        "Dice",
        "Mozley",
        "Ochiai",
        "Simpson",
        "cosine",
        "angular",
        "correlation",
        "Tschuprow",
        "Cramer",
        "Gower",
        "Euclidean",
        "Manhattan",
        "supremum",
        "Canberra",
        "Chord",
        "Geodesic",
        "Bray",
        "Soergel",
        "Podani",
        "Whittaker",
        "eJaccard",
        "eDice",
        "Bhjattacharyya",
        "divergence",
        "Hellinger",
        "edit_sim_utf8",
        "edit_sim",
        "Levenshtein",
        "sim_NCD",
        "const",
        "sim_dtw",
    ]

    supported_transformations = [
        "pitch",
        "int",
        "fuzzy_int",
        "parsons",
        "pc",
        "ioi_class",
        "duration_class",
        "int_X_ioi_class",
        "implicit_harmonies",
    ]

    # Use batch processing to test all combinations efficiently
    results = get_similarities(
        {"melody1": mel_1, "melody2": mel_2},
        method=supported_measures,
        transformation=supported_transformations,
    )

    # Verify we got results for all combinations
    expected_combinations = len(supported_measures) * len(
        supported_transformations
    )
    assert (
        len(results) == expected_combinations
    ), f"Expected {expected_combinations} results, got {len(results)}"

    # Test each result matrix
    for (method, transformation), matrix in results.items():
        assert matrix is not None, f"Failed for {method} with {transformation}"
        assert isinstance(
            matrix, dict
        ), f"Result for {method} with {transformation} is not a dict"

        # Check matrix structure
        assert (
            "melody1" in matrix
        ), f"Missing melody1 in matrix for {method} with {transformation}"
        assert (
            "melody2" in matrix
        ), f"Missing melody2 in matrix for {method} with {transformation}"
        assert (
            "melody1" in matrix["melody2"]
        ), f"Missing melody1 in melody2 row for {method} with {transformation}"
        assert (
            "melody2" in matrix["melody1"]
        ), f"Missing melody2 in melody1 row for {method} with {transformation}"

        # Check diagonal values
        assert (
            matrix["melody1"]["melody1"] == 1.0
        ), f"Diagonal not 1.0 for {method} with {transformation}"
        assert (
            matrix["melody2"]["melody2"] == 1.0
        ), f"Diagonal not 1.0 for {method} with {transformation}"

        # Check symmetry and range
        similarity = matrix["melody1"]["melody2"]
        reverse_similarity = matrix["melody2"]["melody1"]

        # Check if both are NaN (NaN != NaN in Python, so we need special handling)
        both_nan = (
            isinstance(similarity, float)
            and math.isnan(similarity)
            and isinstance(reverse_similarity, float)
            and math.isnan(reverse_similarity)
        )

        # Assert symmetry (treating two NaN values as equal)
        assert (
            both_nan or similarity == reverse_similarity
        ), f"Matrix not symmetric for {method} with {transformation}"

        # Handle cases where similarity is NaN (invalid combinations)
        if not math.isnan(similarity):
            assert isinstance(
                similarity, float
            ), f"Result for {method} with {transformation} is not a float"
            assert (
                0.0 <= similarity <= 1.0
            ), f"Similarity out of range for {method} with {transformation}: {similarity}"


def test_batch_similarity():
    """Test batch similarity calculation."""
    mel_1 = Score.from_melody(pitches=[60, 62, 64, 65], durations=1.0)
    mel_2 = Score.from_melody(pitches=[60, 62, 64, 67], durations=1.0)
    mel_3 = Score.from_melody(pitches=[67, 69, 71, 72], durations=1.0)

    scores = {"melody1": mel_1, "melody2": mel_2, "melody3": mel_3}

    # Test single method/transformation - should return a nested dictionary
    results = get_similarities(scores, method="Jaccard", transformation="pitch")

    # Should be a nested dictionary {melody_name: {melody_name: similarity}}
    assert isinstance(results, dict)
    assert "melody1" in results
    assert "melody2" in results["melody1"]
    assert "melody3" in results["melody1"]

    # Diagonal should be 1.0
    assert results["melody1"]["melody1"] == 1.0
    assert results["melody2"]["melody2"] == 1.0
    assert results["melody3"]["melody3"] == 1.0

    # Should be symmetric
    assert results["melody1"]["melody2"] == results["melody2"]["melody1"]
    assert results["melody1"]["melody3"] == results["melody3"]["melody1"]
    assert results["melody2"]["melody3"] == results["melody3"]["melody2"]

    # Test multiple methods/transformations - should return dict of matrices
    results_multi = get_similarities(
        scores, method=["Jaccard", "Dice"], transformation=["pitch", "int"]
    )

    # Should be a dictionary mapping (method, transformation) tuples to matrices
    assert isinstance(results_multi, dict)
    assert ("Jaccard", "pitch") in results_multi
    assert ("Jaccard", "int") in results_multi
    assert ("Dice", "pitch") in results_multi
    assert ("Dice", "int") in results_multi

    # Each matrix should be a nested dictionary
    jaccard_pitch_matrix = results_multi[("Jaccard", "pitch")]
    assert isinstance(jaccard_pitch_matrix, dict)
    assert "melody1" in jaccard_pitch_matrix
    assert "melody2" in jaccard_pitch_matrix["melody1"]

    # Diagonal should be 1.0
    assert jaccard_pitch_matrix["melody1"]["melody1"] == 1.0



"""functions for file output"""

__author__ = "Roger B. Dannenberg"

import pathlib
import warnings
from typing import Callable, Optional

from amads.core.basics import Score

# This module, writescore, is regarded as a singleton class with
# the following attributes:

preferred_midi_writer: str = "pretty_midi"  # subsystem for MIDI file output
preferred_xml_writer: str = "music21"  # subsystem for MusicXML file output
preferred_kern_writer: str = "pretty_midi"  # subsystem for Kern file output
preferred_mei_writer: str = "pretty_midi"  # subsystem for Kern file output
preferred_pdf_writer: str = "music21-lilypond"  # subsystems for PDF file output
writer_warning_level: str = "default"  # controls verbocity of warnings
#     from output processing
_last_used_writer: Optional[Callable] = None  # See last_used_writer()

# mappings from suffix to format:
_suffix_to_format = {
    ".xml": "musicxml",
    ".musicxml": "musicxml",
    ".mxl": "musicxml",
    ".mid": "midi",
    ".midi": "midi",
    ".smf": "midi",
    ".krn": "kern",
    ".kern": "kern",
    ".mei": "mei",
    ".pdf": "pdf",
    ".ly": "lilypond",
}

# A list of supported file extensions for score writing.
valid_score_extensions = _suffix_to_format.keys()

allowed_subsystems = {
    "midi": ["music21", "pretty_midi"],
    "musicxml": ["music21", "partitura"],
    "kern": ["music21"],
    "mei": ["music21"],
    "pdf": [
        "music21-lilypond",
        "music21-xml-lilypond",
        "partitura-xml-lilypond",
    ],
    "lilypond": [
        "music21-lilypond",
        "music21-xml-lilypond",
        "partitura-xml-lilypond",
    ],
}

# mapping from preference strings to subsystem callables
_subsystem_map = {
    "music21": ("amads.io.m21_export", "music21_export"),
    "pretty_midi": ("amads.io.pm_midi_export", "pretty_midi_export"),
    "partitura": ("amads.io.pt_export", "partitura_export"),
    "music21-lilypond": ("amads.io.m21_pdf_export", "music21_pdf_export"),
    "music21-xml-lilypond": (
        "amads.io.m21_pdf_export",
        "music21_xml_pdf_export",
    ),
    "partitura-xml-lilypond": (
        "amads.io.pt_pdf_export",
        "partitura_xml_pdf_export",
    ),
}


def set_preferred_midi_writer(writer: str) -> str:
    """Set a (new) preferred MIDI writer.

    Returns the previous writer preference. The current preference is stored
    in `amads.io.writer.preferred_midi_writer`.

    Parameters
    ----------
    writer : str
        The name of the preferred MIDI writer. Can be "music21" or "pretty_midi".

    Returns
    -------
    str
        The previous name of the preferred MIDI writer.

    Raises
    ------
    ValueError
        If an invalid writer is provided.

    """
    global preferred_midi_writer
    previous_writer = preferred_midi_writer
    if writer in ["music21", "partitura", "pretty_midi"]:
        preferred_midi_writer = writer
    else:
        raise ValueError(
            "Invalid MIDI writer. Choose 'music21', 'partitura', or "
            "'pretty_midi'."
        )
    return previous_writer


def set_preferred_xml_writer(writer: str) -> str:
    """
    Set a (new) preferred XML writer.

    Returns the previous writer preference. The current preference is stored
    in `amads.io.writer.preferred_xml_writer`.

    Parameters
    ----------
    writer : str
        The name of the preferred XML writer. Can be "music21" or "partitura".

    Returns
    -------
    str
        The previous name of the preferred XML writer.

    Raises
    ------
    ValueError
        If an invalid writer is provided.
    """
    global preferred_xml_writer
    previous_writer = preferred_xml_writer
    if writer in allowed_subsystems["musicxml"]:
        preferred_xml_writer = writer
    else:
        raise ValueError("Invalid XML writer. Choose 'music21' or 'partitura'.")
    return previous_writer


def set_preferred_kern_writer(writer: str) -> str:
    """
    Set a (new) preferred Kern writer.

    Returns the previous writer preference. The current preference is stored
    in `amads.io.writer.preferred_kern_writer`.

    Parameters
    ----------
    writer : str
        The name of the preferred Kern writer. Can be "music21".

    Returns
    -------
    str
        The previous name of the preferred Kern writer.

    Raises
    ------
    ValueError
        If an invalid writer is provided.
    """
    global preferred_kern_writer
    previous_writer = preferred_kern_writer
    if writer in allowed_subsystems["kern"]:
        preferred_kern_writer = writer
    else:
        raise ValueError("Invalid Kern writer. Choose 'music21'.")
    return previous_writer


def set_preferred_mei_writer(writer: str) -> str:
    """
    Set a (new) preferred MEI writer.

    Returns the previous writer preference. The current preference is stored
    in `amads.io.writer.preferred_mei_writer`.

    Parameters
    ----------
    writer : str
        The name of the preferred MEI writer. Can be "music21".

    Returns
    -------
    str
        The previous name of the preferred MEI writer.

    Raises
    ------
    ValueError
        If an invalid writer is provided.
    """
    global preferred_mei_writer
    previous_writer = preferred_mei_writer
    if writer in allowed_subsystems["mei"]:
        preferred_mei_writer = writer
    else:
        raise ValueError("Invalid MEI writer. Choose 'music21'.")
    return previous_writer


def set_preferred_pdf_writer(writer: str) -> str:
    """
    Set a (new) preferred PDF writer.

    Returns the previous writer preference. The current preference is stored
    in `amads.io.writescore.preferred_pdf_writer`. Preferences are:
    - "music21-lilypond" - use music21 to create a LilyPond file, then use
      LilyPond to create a PDF.
    - "music21-xml-lilypond" - use music21 to create a MusicXML file, then run
      the program musicxml2ly to convert XML to LilyPond, then run LilyPond to
      create a PDF.
    - "partitura-xml-lilypond" - use partitura to create a MusicXML file, then
      run the program musicxml2ly to convert XML to LilyPond, then run LilyPond
      to create a PDF.


    Parameters
    ----------
    writer : str
        The name of the preferred PDF writer. Can be "music21-lilypond",
        "music21-xml-lilypond", or "partitura-xml-lilypond".

    Returns
    -------
    str
        The previous name of the preferred PDF writer.

    Raises
    ------
    ValueError
        If an invalid writer is provided.
    """
    global preferred_pdf_writer
    previous_writer = preferred_pdf_writer
    if writer in allowed_subsystems["pdf"]:
        preferred_pdf_writer = writer
    else:
        raise ValueError(
            "Invalid PDF writer. Choose " f"{allowed_subsystems['pdf']}."
        )
    return previous_writer


def set_writer_warning_level(level: str) -> str:
    """
    Set the warning level for writescore functions.

    The translation from AMADS to music data files is not always well-defined
    and may involve intermediate representations using Music21, Partitura or
    others. Usually, warnings are produced when there is possible data loss or
    ambiguity, but these can be more annoying than informative. The warning
    level can be controlled using this function, which applies to all file
    formats.

    The current warning level is stored in
    `amads.io.writer.writer_warning_level`.

    Parameters
    ----------
    level : str
        The warning level to set. Can be "none", "low", "default", "high".

        - "none" - will suppress all warnings during write_score().
        - "low" - will show print one notice if there are any warnings.
        - "default" - will obey environment settings to control warnings.
        - "high" - will print all warnings during write_score(), overriding
            environment settings.

    Returns
    -------
    str
        Previous warning level.

    Raises
    -------
    ValueError
        If an invalid warning level is provided.
    """
    global writer_warning_level
    previous_level = writer_warning_level
    if level in ["none", "low", "default", "high"]:
        writer_warning_level = level
    else:
        raise ValueError(
            "Invalid warning level. Choose 'none', 'low', 'default', or 'high'."
        )
    return previous_level


def _check_for_subsystem(
    format: str,
) -> tuple[
    Optional[Callable[[Score, Optional[str], Optional[str], bool, bool], None]],
    Optional[str],
]:
    """Check if the preferred subsystem is available.

    Parameters
    ----------
    format : str
        The format of the file to write, either 'midi', 'musicxml', or 'pdf'.

    Returns
    -------
    tuple[Optional[Callable], Optional[str]]
        The export function if available, None otherwise, and the name of
        the subsystem used.
    """
    preferred_writer = {
        "midi": preferred_midi_writer,
        "musicxml": preferred_xml_writer,
        "kern": preferred_midi_writer,
        "mei": preferred_midi_writer,
        "pdf": preferred_pdf_writer,
    }.get(format)

    if not preferred_writer:
        return None, None

    try:
        if (
            preferred_writer not in _subsystem_map
            or preferred_writer not in allowed_subsystems[format]
        ):
            raise ValueError(
                f"Preferred writer '{preferred_writer}' not supported for "
                f"{format} export."
            )

        module_name, func_name = _subsystem_map[preferred_writer]
        module = __import__(module_name, fromlist=[func_name])
        # note that the type signature of func_name has an extra `display``
        # when format is "pdf", so if you call it with a `display` argument,
        # type checking will complain:
        return getattr(module, func_name), preferred_writer
    except Exception as e:
        print(f"Error importing {preferred_writer} for {format} files: {e}")
    return None, preferred_writer


def _export_score(
    score: Score,
    filename: Optional[str],
    format: str,
    show: bool = False,
    display: bool = False,
) -> None:
    """Use Partitura or music21 to export a MusicXML file.

    <small>**Author**: Roger B. Dannenberg</small>
    """
    global _last_used_writer

    export_fn, subsystem = _check_for_subsystem(format)
    if export_fn is not None:
        _last_used_writer = export_fn
        if writer_warning_level != "none":
            print(
                f"Exporting {filename} using {format} writer"
                f" {export_fn.__name__} from subsystem {subsystem}."
            )
        export_fn(score, filename, format, show, display)
    else:
        raise Exception(
            f"Could not find an export function for format {format}. "
            "Preferred subsystem is " + str(subsystem)
        )


def write_score(
    score: Score,
    filename: str,
    show: bool = False,
    format: Optional[str] = None,
) -> None:
    """Write a file with the given format.

    If format is None (default), the format is based on the filename
    extension, which can be one of `writescore.valid_score_extensions`
    ('xml', 'musicxml', 'mxl', 'mid', 'midi', 'smf', 'krn', 'kern',
    'mei', 'pdf', or 'ly').

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    score : Score
        the score to write
    filename : Optional[str]
        the path (relative or absolute) to the music file. Optional only
        when display is True, but for display, you should call `display_score`
        instead.
    show : bool
        print a text representation of the data
    format : Optional[string]
        one of `'musicxml'`, `'midi'`, `'kern'`, `'mei'`, `'pdf'`, `'lilypond'`.
        Defaults to the format implied by `filename`.

    Raises
    ------
    ValueError
        if format is unknown

    Notes
    -----
    AMADS assumes that instruments (midi program numbers) are fixed
    for each Staff (or Part in flat scores), and MIDI channels are
    not represented. This corresponds to some DAWs such as LogicPro,
    which represents channels but ignores them when tracks are
    synthesized in software by a single instrument. The MIDI program
    is stored as info (see [get][amads.core.basics.Event.get] and
    [set][amads.core.basics.Event.set]) under key `"midi_program"`
    on the Staff, or if there is no Staff or no `"midi_program"` on
    the Staff, under key `"midi_program"` on the Part.

    Parts also have an `instrument` attribute, which is stored as
    the MIDI track name. (Therefore, if a Part has two Staffs, there
    will be two tracks with the same name.)  If there is no MIDI
    program for the track, the `'pretty_midi'` writer will use
    `pretty_midi.instrument_name_to_program` to determine a program
    number since a program number is required. (As opposed to Standard
    MIDI Files, which need not have any MIDI program message at all.)
    If `pretty_midi.instrument_name_to_program` fails, the program is
    set to 0 (“Acoustic Grand Piano”).

    Partitura does not seem to support per-staff key signatures,
    so key signatures from AMADS are simply added to Partitura
    parts. When there are multiple staffs, there could be
    duplicate key signatures (to be tested).

    Pretty MIDI also requires an instrument name. If the AMADS Part
    `instrument` attribute is `None`, then `"Unknown"` is used. The
    Pretty MIDI reader will convert `"Unknown"` back to `None`.

    """
    _write_or_display_score(score, filename, show, format, False)


def _write_or_display_score(
    score: Score,
    filename: Optional[str],
    show: bool = False,
    format: Optional[str] = None,
    display: bool = False,
) -> None:
    """Write or display a Score.

    If format is None (default), the format is based on the filename
    extension, which can be one of `writescore.valid_score_extensions`
    ('xml', 'musicxml', 'mxl', 'mid', 'midi', 'smf', 'krn', 'kern',
    'mei', 'pdf', or 'ly').

    If display is True, the goal is to display the file, so the
    filename is optional, and a temporary file will be used if needed.

    display is suppressed by setting AMADS_NO_OPEN=1 in the environment,
    which is used in testing with pytest so the user does not have to close
    a bunch of windows opened by demos that are run as part of testing.

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    score : Score
        the score to write
    filename : Optional[str]
        the path (relative or absolute) to the music file. Optional only
        when display is True, but for display, you should call `display_score`
        instead.
    show : bool
        print a text representation of the data
    format : Optional[string]
        one of 'musicxml', 'midi', 'kern', 'mei', 'pdf', 'lilypond'.
        Defaults to the format implied by `filename`.
    display : bool
        If True and format is 'pdf', the created PDF file is displayed.

    Raises
    ------
    ValueError
        if format is unknown
    """

    if not display and not filename:
        raise ValueError("filename must be provided if display is False")
    if format is None and not filename:
        raise ValueError("format must be provided if filename is not provided")

    # Type checking complains about filename being possibly None, but we
    # check for that above, so we can ignore it here.
    if format is None and filename:
        ext = pathlib.Path(filename).suffix  # type: ignore
        format = _suffix_to_format.get(ext)
        if not format:
            raise ValueError(
                f"Unsupported file extension: {ext}. "
                f"Valid extensions: {valid_score_extensions}"
            )
    elif format not in _suffix_to_format.values():
        raise ValueError(f"Unknown or unspecified format: {format}")

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter(
            "ignore" if writer_warning_level == "none" else "always"
        )

        # here, we know format is valid, so ignore type checking:
        _export_score(score, filename, format, show, display)  # type: ignore

        # Warning handling
        if writer_warning_level == "low":
            if len(w) > 0:
                print(
                    f"Warning: {len(w)} warnings were generated in"
                    f" write_score({filename}). Use"
                    " amads.io.writescore.set_writer_warning_level() for"
                    " more details."
                )
        else:  # "none", "default", or "high"
            for warning in w:
                formatted = warnings.formatwarning(
                    warning.message,
                    warning.category,
                    warning.filename,
                    warning.lineno,
                )
                print(formatted, end="")



import numpy as np
import pytest

import amads.pitch.key.profiles as prof
from amads.core.basics import Score
from amads.io.readscore import read_score
from amads.music import example
from amads.pitch.key.key_cc import key_cc


def test_error_handling():
    """Invalid attribute name should return None"""
    score = Score.from_melody([60, 62, 64])
    result = key_cc(score, prof.krumhansl_kessler, ["invalid_attribute"])
    assert result == [("invalid_attribute", None)]


def test_empty_melody():
    score = Score.from_melody([])
    with pytest.raises(RuntimeError):
        key_cc(score, prof.krumhansl_kessler, ["major", "minor"])


def test_equal_prob_melody():
    """
    Equal prob melody containing 1 of each pitch.
    Without salience this should return all 24 coefficients as equivalent
    for all 3 profiles of kkcc
    """
    melody = list(range(60, 72))  # C4 to B4
    score = Score.from_melody(melody)
    for profile in [
        prof.krumhansl_kessler,
        prof.temperley,
        prof.albrecht_shanahan,
    ]:
        with pytest.raises(RuntimeError):
            key_cc(score, profile, ["major", "minor"])
    return


def test_none_input():
    score = Score.from_melody([60, 62, 64])
    result = key_cc(score, prof.krumhansl_kessler)
    names = [name for name, _ in result]
    # _sum options removed, len(result) changed from 4 to 2
    assert "major" in names and "minor" in names and len(result) == 2
    for _, corr in result:
        assert corr is None or (len(corr) == 12)


def test_crafted_nonempty_melodies():
    """
    These melodies are here to test the various codepaths and are specific
    to the implementation itself...
    """
    # C major scale
    score = Score.from_melody([60, 62, 64, 65, 67, 69, 71, 72])

    # transpositionally equivalent
    result = key_cc(score, prof.bellman_budge, ["major", "minor"])
    assert isinstance(result, list)
    assert len(result) == 2
    for (attr, corr), expected in zip(result, ["major", "minor"]):
        assert attr == expected
        assert corr is not None
        assert len(corr) == 12
        assert all(isinstance(x, float) for x in corr)  # floats expected

    # transpositionally unequivalent
    result2 = key_cc(score, prof.quinn_white, ["major_asym", "minor_asym"])
    assert isinstance(result2, list)
    assert len(result2) == 2
    for (attr, corr), expected in zip(result2, ["major_asym", "minor_asym"]):
        assert attr == expected
        assert corr is not None
        assert len(corr) == 12
        assert all(isinstance(x, float) for x in corr)  # floats expected


def test_salience():
    score = Score.from_melody([60, 62, 64, 65, 67, 69, 71, 72])
    result = key_cc(
        score, prof.bellman_budge, ["major", "minor"], salience_flag=True
    )

    desired_result = [
        (
            "major",
            (
                0.91269250175412,
                -0.7487980568570414,
                0.3831455804244438,
                -0.2403214013870691,
                -0.12991952699984174,
                0.6151931033572154,
                -0.8619783953942095,
                0.7959468120362349,
                -0.44534607214102767,
                0.11127537312370366,
                0.1414059123565155,
                -0.5332958302730434,
            ),
        ),
        (
            "minor",
            (
                0.5122258362978889,
                -0.4512145246564831,
                0.3454593728930353,
                -0.6159365356889505,
                0.46487216789241365,
                0.15021417133013423,
                -0.4536556405976428,
                0.42387649045308,
                -0.5364189966155352,
                0.5830992373851366,
                -0.33248273987015997,
                -0.09003883882291706,
            ),
        ),
    ]
    for (attr, corr), (expected_attr, expected_corr) in zip(
        result, desired_result
    ):
        assert attr == expected_attr
        assert np.allclose(corr, expected_corr, rtol=1e-15)


def test_sarabande():
    from amads.pitch.pcdist1 import pitch_class_distribution_1

    my_midi_file = example.fullpath("midi/sarabande.mid")
    myscore = read_score(my_midi_file, show=False)

    pcd = np.array([pitch_class_distribution_1(myscore, False)])
    print("Pitch-Class Distribution:", pcd)

    result = key_cc(myscore, prof.krumhansl_kessler, ["major", "minor"])

    assert isinstance(result, list)
    assert len(result) == 2
    for (attr, corr), expected in zip(result, ["major", "minor"]):
        assert attr == expected
        assert corr is not None
        assert len(corr) == 12
        assert all(isinstance(x, float) for x in corr)  # floats expected

    desired_result = [
        (
            "major",
            (
                0.618783892738462,
                -0.6519513221653466,
                0.4875907439536567,
                -0.5290638735709131,
                0.2542559399388133,
                0.4379780281498363,
                -0.7149269867198986,
                0.5069337968414559,
                -0.45268834604256136,
                0.44440110296372387,
                -0.15309251162008541,
                -0.24822046446714316,
            ),
        ),
        (
            "minor",
            (
                -0.044203856666773005,
                -0.1307611916483344,
                0.5377973285450691,
                -0.7181318384052082,
                0.4801058958544669,
                -0.07985223669882462,
                -0.022968241834542897,
                -0.13552507482362564,
                -0.3137266030925678,
                0.7957737140544804,
                -0.6323415693586513,
                0.2638336740745113,
            ),
        ),
    ]

    for (attr, corr), (expected_attr, expected_corr) in zip(
        result, desired_result
    ):
        assert attr == expected_attr
        assert np.allclose(corr, expected_corr, rtol=1e-15)


def test_random_generated_melodies():
    """
    These are randomly generated (but deterministic due to fixed seed) melodies
    Are we sure we have a reference implementation or a decent set of
    verification properties to test against to pull this one off?
    """
    np.random.seed(42)
    test_iterations = 10
    for _ in range(test_iterations):
        pitches = np.random.randint(
            60, 72, size=16
        )  # length 16, pitches from C4 to B4
        score = Score.from_melody(pitches.tolist())
        result = key_cc(score, prof.sapp, ["major", "minor"])
        assert isinstance(result, list)
        assert len(result) == 2
        for (attr, corr), expected in zip(result, ["major", "minor"]):
            assert attr == expected
            assert corr is not None
            assert len(corr) == 12
            assert all(isinstance(x, float) for x in corr)  # floats expected



"""Test suite for key profiles in `pitch.key.profiles.py`"""

from dataclasses import fields

import pytest

from amads.pitch.key.profiles import PitchProfile, source_list


def test_string_attributes():
    """Every key profile should define `name`, `literature`, and `about` attributes as (non-empty) strings"""
    expected_attrs = ["name", "literature", "about"]
    for profile in source_list:
        # This dunder method should just be set to the name of the profile
        assert str(profile()) == profile().__str__() == getattr(profile, "name")
        # The name of the class should be the same as its .name attribute
        assert profile().__class__.__name__ == getattr(profile, "name")
        # The class should have all the desired attributes as non-empty strings
        for attr in expected_attrs:
            assert hasattr(profile, attr)
            assert isinstance(getattr(profile, attr), str)
            assert getattr(profile, attr) != ""


def test_data_attributes():
    """Check attributes provided for each dataclass"""
    for profile in source_list:
        attribute_names = [f.name for f in fields(profile)]
        # Iterate over every attribute for this profile
        for attr_key in attribute_names:
            # Skip over this attribute, which gives us a tuple
            if attr_key == "__match_args__":
                continue
            # Get the values of attribute
            attr_val = getattr(profile, attr_key)
            # Skip over reserved metadata attribute names
            if attr_key in ["name", "literature", "about"]:
                assert isinstance(attr_val, str)
            # Every element of tuples should be a floating point number
            #  We should have 12 of them, one per key
            elif isinstance(attr_val, tuple):
                assert len(attr_val) == 12
                for elem in attr_val:
                    if not isinstance(elem, PitchProfile):
                        print(
                            f"Attribute {attr_key} of profile {profile.__name__} "
                            f"has non-PitchProfile element: {elem} ({type(elem)})"
                        )
                assert all(isinstance(elem, PitchProfile) for elem in attr_val)
            else:
                assert isinstance(attr_val, PitchProfile)


def test_missing_attributes():
    """Test that we raise errors properly when trying to access missing attributes"""
    for profile in source_list:
        with pytest.raises(AttributeError):
            _ = profile().__getitem__("missing")



"""
pm_midi_import.py

Import MIDI files into AMADS Score structure using the PrettyMIDI library.

Functions
---------
- pretty_midi_midi_import(filename: str, flatten: bool = False,
                          collapse: bool = False, show: bool = False) -> Score:
    Imports a MIDI file using PrettyMIDI and converts it into a `Score` object.

Usage
-----
Do not use this module directly; see readscore.py.

Notes
-----
PrettyMIDI has a "hidden" representation of the MIDI tempo track in
`_tick_scales`, which has the form [(tick, tick_duration), ...]. We
need to convert this to a TimeMap.
"""

__author__ = "Roger B. Dannenberg"

import warnings
from math import isclose
from typing import cast

from pretty_midi import PrettyMIDI, program_to_instrument_name

# this should not be necessary, but Python confuses AMADS with
#     pretty_midi's TimeSignature class:
from amads.core.basics import (
    KeySignature,
    Measure,
    Note,
    Part,
    Score,
    Staff,
    TimeSignature,
)
from amads.core.pitch import Pitch
from amads.core.timemap import TimeMap


def _time_map_from_tick_scales(tick_scales, resolution: int) -> TimeMap:
    """Convert "hidden" _tick_scales list to AMADS TimeMap, representing the
    midi file tempo map
    """
    ticks_per_second = 1.0 / tick_scales[0][1]
    time_map = TimeMap(qpm=ticks_per_second * 60 / resolution)
    for change in tick_scales[1:]:
        # tick_scales is not documented, but since this works, we
        # can say that tick_scales is a list of breakpoints where
        # tempo changes. time (sec) = change[0] / resolution, and
        # beat (quarters) = 60 / (change[1] * resolution), so
        # each tuple is (score time in units of ticks, seconds/tick)
        time_map.append_change(
            change[0] / resolution, 60.0 / (change[1] * resolution)
        )
    return time_map


def _create_measures(
    staff: Staff,
    time_map: TimeMap,
    pmscore: PrettyMIDI,
) -> None:
    """Create measures in Staff according to pmscore time_signature_changes

    At each iteration, insert measures up to the signature time, and
    on the last iteration, insert measures up to the end of the score

    To deal with float approximation, we round beat times to 1/32 since
    time signatures are always n/(2^d), e.g. 4/4, 3/16, ....
    """
    # insert measures according to time_signature_changes:
    # also add key signatures to measures
    cur_upper = 4
    cur_lower = 4
    cur_duration = 4  # default is 4/4 starting at beat 0
    cur_beat = 0  # we have created measures up to this time

    ksigs = pmscore.key_signature_changes
    k = 0  # index to key signatures in ksigs
    if k < len(ksigs):
        ksig = ksigs[k]
        kbeat = time_map.time_to_quarter(ksig.time)
    else:
        kbeat = staff.duration + 999  # no key signatures

    tsigs = pmscore.time_signature_changes
    i = 0
    score = cast(Score, staff.score)
    while i <= len(tsigs):
        if i < len(tsigs):
            tsig = tsigs[i]
            tbeat = time_map.time_to_quarter(tsig.time)
        else:  # trick to fill measures to the end of score
            tsig = None
            tbeat = score.duration  # should we add 1e-6 here?
        need_time_signature = True
        while cur_beat < tbeat - 1e-6:  # avoid rounding
            measure = Measure(onset=cur_beat, duration=cur_duration)
            if need_time_signature:  # first measure after time signature change
                ts = TimeSignature(cur_beat, cur_upper, cur_lower)
                score.append_time_signature(ts)
                need_time_signature = False
            staff.insert(measure)
            # if needed now, insert key signature into measure
            if cur_beat > kbeat - 1e-6:  # avoid rounding error
                _ = KeySignature(
                    measure, cur_beat, ksig.key_number  # type: ignore
                )  # type: ignore
                # get next key signature
                k += 1
                if k < len(ksigs):
                    ksig = ksigs[k]
                    kbeat = time_map.time_to_quarter(ksig.time)
                else:
                    kbeat = score.duration + 999  # no more key signatures
            cur_beat += cur_duration

        # now set up for next iteration
        if tsig is not None:
            tsig_str = str(tsig.numerator) + "/" + str(tsig.denominator)
            if not isclose(cur_beat, tbeat, abs_tol=1e-6):
                warnings.warn(
                    f"MIDI file time signature change at beat {tbeat}"
                    " is not on the expected measure boundary at"
                    f" {cur_beat}. The time signature {tsig_str}"
                    f" will be applied at {cur_beat}."
                )
            cur_upper = tsig.numerator
            cur_lower = tsig.denominator
            cur_duration = cur_upper * 4 / cur_lower
        # else tsig is None means we've reached the end of the score
        i += 1


def _add_notes_to_measures(
    notes: list[Note], measures: list[Measure], div: int
) -> None:
    """Add notes to measures and tie notes across measure boundaries.

    Parameters
    ----------
    notes : list[Note]
        The notes to insert. Parent is set in these notes, but the
        notes are not in the parent's content. Parent will be reset
        to the measure when the note is inserted.
    measures: list[Measures]
        The measures were the notes go.
    div: int
        The original number of divisions per quarter, used to perform
        rounding.
    """
    EPS = 0.5 / div  # time resolution ("epsilon")
    i = 0  # notes index
    for mi, m in enumerate(measures):
        m_insert = m  # where to insert note
        m_insert_i = mi  # index of measure to insert into
        while i < len(notes) and notes[i].onset < m.offset:
            note = notes[i]
            note.parent = None  # fix incorrect parent attribute
            # break note if it spans a measure boundary
            if note.onset > m.offset - EPS:
                # note is at the very end of the measure; just round up
                offset = note.offset
                m_insert_i = mi + 1  # index of measure to insert into
                m_insert = measures[m_insert_i]
                note.onset = m_insert.onset
                note.duration = offset - note.onset  # don't change offset time
            remaining = 0
            if note.offset > m_insert.offset:  # split and tie note
                remaining = note.offset - m_insert.offset
                note.duration = m_insert.offset - note.onset
            m_insert.insert(note)

            next_i = m_insert_i + 1
            prev_note = note
            while remaining > EPS:
                next_measure = measures[next_i]
                duration = min(remaining, next_measure.duration)
                tied_note = Note(
                    parent=next_measure,
                    onset=next_measure.onset,
                    duration=duration,
                    pitch=Pitch(note.pitch),
                    dynamic=note.dynamic,
                )
                prev_note.tie = tied_note
                prev_note = tied_note
                next_i += 1
                remaining -= tied_note.duration
            i += 1


# debug
def has_staff(part):
    for item in part.content:
        if isinstance(item, Staff):
            return True
    return False


def pretty_midi_import(
    filename: str,
    format: str,
    flatten: bool = False,
    collapse: bool = False,
    show: bool = False,
    group_by_instrument: bool = True,
) -> Score:
    """
    Use PrettyMIDI to import a MIDI file and convert it to a Score.

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    filename : Union(str, PosixPath)
        The path to the MIDI file to import.
    format : str
        The format of the MIDI file. Must be "midi".
    flatten : bool, optional
        If True, create a flat score where notes are direct children of
        parts. Defaults to collapse, which defaults to False.
    collapse : bool, optional
        If True, merge all parts into a single part. Implies flatten=True.
        Defaults to False.
    show : bool, optional
        If True, print the PrettyMIDI score structure for debugging.
        Defaults to False.
    group_by_instrument : bool, optional
        If True, group parts by instrument name into staffs. Defaults to True.
        See read_midi() for more details.

    Returns
    -------
    Score
        The converted Score object containing the imported MIDI data.

    Examples
    --------
    >>> from amads.io.pm_midi_import import pretty_midi_import
    >>> from amads.music import example
    >>> score = pretty_midi_import( \
                    example.fullpath("midi/sarabande.mid"), "midi",\
                    flatten=True)  # show=True to see PrettyMIDI data
    """
    flatten = flatten or collapse  # collapse implies flatten

    # Load the MIDI file using PrettyMidi
    filename = str(filename)
    pmscore = PrettyMIDI(filename)
    if show:
        from amads.io.pm_show import pretty_midi_show

        pretty_midi_show(pmscore, filename)

    # Create an empty Score object
    time_map = _time_map_from_tick_scales(
        pmscore._tick_scales, pmscore.resolution
    )
    score = Score(time_map=time_map)
    score.convert_to_seconds()  # convert to seconds for PrettyMIDI

    # Iterate over instruments of the PrettyMIDI score and build parts and notes
    # Then if collapse, merge and sort the notes
    # Then if not flatten, remove each part content, and staff and measures,
    # and move notes into measures, creating ties where they cross

    # This helps orgainize parts by instrument name. Parts with the same
    # instrument name are grouped together as staffs in a single Part.
    instrument_groups: dict[str, list[Part]] = {}

    # Iterate over instruments of the PrettyMIDI score and build parts and notes
    duration = 0
    no_name_id = 1
    for ins in pmscore.instruments:
        name = ins.name
        if not ins.name:
            name = program_to_instrument_name(ins.program)
        if name == "Unknown":  # AMADS uses "Unknown" to represent None.
            # Of course, if a user really names an instrument "Unknown",
            # that particular name will not be stored in the Part.
            name = None

        part = Part(parent=score, onset=0.0, instrument=name)

        # now that we've put the name in the Part, change None to a unique name
        # so that we can make the part a named instrument group. If we are not
        # grouping by instrument, we want each part in its own group, so we
        # give each part a unique name in that case too.
        if name is None or not group_by_instrument:
            name = f"_None~@_{no_name_id}"  # something unlikely to be a name
            no_name_id += 1
        group = instrument_groups.get(name, [])
        group.append(part)
        instrument_groups[name] = group

        for note in ins.notes:
            # Create a Note object and associate it with the Part
            Note(
                parent=part,
                onset=note.start,
                duration=note.get_duration(),
                pitch=Pitch(note.pitch),
                dynamic=note.velocity,
            )
            duration = max(duration, note.end)
        assert not has_staff(
            part
        ), "Part should not have Staffs before flattening"

    score.duration = duration
    for part in score.content:
        part.duration = duration  # all parts get same max duration
        assert not has_staff(part), "Part should not have Staffs"

    # Then if collapse, merge and sort the notes
    if collapse:
        score = score.flatten(collapse=True)

    score.convert_to_quarters()  # we want to return with quarters as time unit

    # TODO: remove this block
    for groups in instrument_groups.values():
        for part in groups:
            assert not has_staff(part), "Part should not have Staffs 2"

    # Then if not flatten, remove each part content, and staff and measures,
    # and move notes into measures, creating ties where they cross. This maybe
    # does some extra work if not grouping by instrument, because we already
    # have the right set of parts, but the real work here is creating the staffs
    # and measures, and moving notes into measures with ties, so we have
    # eliminated the small extra work of copying parts when not grouping.
    if not flatten:
        score.content.clear()  # remove parts from score
        for group in instrument_groups.values():
            new_part = group[0].insert_emptycopy_into(score)
            new_part = cast(Part, new_part)
            new_part.duration = score.duration
            for old_part in group:
                assert not has_staff(
                    old_part
                ), "Part should not have Staffs before creating staffs"
                old_part = cast(Part, old_part)
                notes = old_part.content
                # now notes have part as parent, but parent does not have notes
                staff = Staff(
                    parent=new_part, onset=0.0, duration=new_part.duration
                )
                # in principle we could do this once for the first staff and
                # then copy the created staff with measures for any other
                # staff, but then we would have to save off the notes and
                # write another loop to insert each note list to a corresponding
                # staff. Besides, _create_measures might even be faster than
                # calling deepcopy on a Staff to copy the measures.
                _create_measures(staff, score.time_map, pmscore)
                notes = cast(
                    list[Note], notes
                )  # tell type checker notes is list of Note
                _add_notes_to_measures(
                    notes,
                    cast(list[Measure], staff.content),
                    pmscore.resolution,
                )
                old_part.content.clear()  # clean up

    return score



import math

import pytest

from amads.core.basics import Score
from amads.melody.fantastic import (
    fantastic_count_mtypes,
    fantastic_huron_contour_features,
    fantastic_interpolation_contour_features,
    fantastic_mtype_summary_features,
    fantastic_parsons_contour_features,
    fantastic_pitch_features,
    fantastic_pitch_interval_features,
    fantastic_polynomial_contour_features,
    fantastic_step_contour_features,
)

__author__ = "David Whyatt"


def test_fantastic_count_mtypes():
    melody = Score.from_melody(
        pitches=[56, 58, 61, 58, 65, 65, 63],
        durations=[0.25, 0.25, 0.25, 0.25, 0.75, 0.75, 0.5],
    )

    types = fantastic_count_mtypes(
        melody, segment=False, phrase_gap=1.0, units="quarters"
    )
    # FANTASTIC supports n-grams of lengths 1-5, so we check that we have n-grams of lengths 1-5
    for i in range(1, 6):
        assert any(len(ngram) == i for ngram in types.ngram_counts.keys())


def test_fantastic_interpolation_contour_features():
    # Test with a melody that generally increases in pitch
    melody = Score.from_melody(
        pitches=[56, 58, 61, 58, 65, 65, 63],
        durations=[0.25, 0.25, 0.25, 0.25, 0.75, 0.75, 0.5],
    )
    features = fantastic_interpolation_contour_features(melody)
    assert features is not None

    # The melody tends to increase in pitch over time,
    # so we expect the global direction to be positive
    assert features["global_direction"] > 0

    # All of the changes in pitch are larger than a semitone
    assert features["mean_gradient"] > 1

    # The standard deviation of the gradient ought to be small
    # as we don't expect much variation
    assert features["gradient_std"] < 0.1

    # By virtue of how FANTASTIC identifies turning points,
    # the direction changes should be 0.0
    assert features["direction_changes"] == 0.0

    # The class label should be dddd
    # as the interpolated contour goes upwards across all four sampled steps
    assert features["class_label"] == "dddd"

    # Test with a descending melody
    descending = Score.from_melody(
        pitches=[67, 65, 64, 62, 60], durations=[1.0, 1.0, 1.0, 1.0, 1.0]
    )
    desc_features = fantastic_interpolation_contour_features(descending)

    # Should have negative global direction
    assert desc_features["global_direction"] < 0
    # Consistent downward motion means low gradient std
    assert desc_features["gradient_std"] < 0.1
    # No direction changes
    assert desc_features["direction_changes"] == 0.0
    # Contour is shallow, so it is classified as cccc
    assert desc_features["class_label"] == "cccc"

    # Test with a flat melody
    flat = Score.from_melody(
        pitches=[60, 60, 60, 60, 60], durations=[1.0, 1.0, 1.0, 1.0, 1.0]
    )
    flat_features = fantastic_interpolation_contour_features(flat)

    # A flat melody should have:
    assert flat_features["global_direction"] == 0  # No overall direction
    assert flat_features["mean_gradient"] == 0.0  # No gradient
    assert flat_features["gradient_std"] == 0.0  # No variation
    assert flat_features["direction_changes"] == 0.0  # No changes
    assert flat_features["class_label"] == "cccc"  # All same


def test_fantastic_step_contour_features():
    melody = Score.from_melody(
        pitches=[56, 58, 61, 58, 65, 65, 63],
        durations=[0.25, 0.25, 0.25, 0.25, 0.75, 0.75, 0.5],
    )

    features = fantastic_step_contour_features(melody)
    assert features is not None

    # The melody has several changes in direction
    # so we expect the global variation to be high
    assert features["global_variation"] > 1.0

    # The melody tends to increase in pitch over time,
    # so we expect the global direction to be positive
    assert features["global_direction"] > 0

    # The melody has several local changes in direction,
    # but not by large intervals, so we expect the local variation to be low
    assert features["local_variation"] < 0.5

    # Test with a simpler melody that only goes up
    simple_melody = Score.from_melody(
        pitches=[60, 62, 64, 65, 67], durations=[1.0, 1.0, 1.0, 1.0, 1.0]
    )

    simple_features = fantastic_step_contour_features(simple_melody)

    # A simple ascending melody should have:
    # Low global variation (consistent upward motion)
    assert simple_features["global_variation"] > 0.5
    # Positive global direction (ascending)
    assert simple_features["global_direction"] > 0
    # Low local variation (no direction changes)
    assert simple_features["local_variation"] < 0.5

    # Test with a melody that stays on same pitch
    flat_melody = Score.from_melody(
        pitches=[60, 60, 60, 60, 60], durations=[1.0, 1.0, 1.0, 1.0, 1.0]
    )

    flat_features = fantastic_step_contour_features(flat_melody)

    # A melody with no pitch changes should have:
    # Zero global variation
    assert flat_features["global_variation"] == 0
    # Zero global direction
    assert flat_features["global_direction"] == 0
    # Zero local variation
    assert flat_features["local_variation"] == 0


def test_fantastic_parsons_contour_features():
    melody = Score.from_melody(
        pitches=[60, 62, 64, 65, 67, 72], durations=[1.0] * 6
    )
    features = fantastic_parsons_contour_features(melody)
    assert features is not None

    # The interval sequence should be [None, 2, 2, 1, 2, 5]
    assert features["interval_sequence"] == [None, 2, 2, 1, 2, 5]

    # As such, the sign sequence would be [None, +1, +1, +1, +1, +1]
    assert features["interval_sequence_sign"] == [None, 1, 1, 1, 1, 1]

    # The string representation should be "uuuuu"
    assert features["as_string"] == "uuuuu"

    # We can also use the initial asterisk option
    features = fantastic_parsons_contour_features(melody, initial_asterisk=True)

    # The string representation should be "*uuuuu"
    assert features["as_string"] == "*uuuuu"

    # We can also use a custom character dict
    features = fantastic_parsons_contour_features(
        melody, character_dict={1: "^", 0: "->", -1: "v"}
    )

    # The string representation should be "^^^^^"
    assert features["as_string"] == "^^^^^"


def test_fantastic_polynomial_contour_features():
    melody = Score.from_melody(
        pitches=[60, 62, 64, 65, 67, 72], durations=[1.0] * 6
    )
    features = fantastic_polynomial_contour_features(melody)
    assert features is not None

    # We would expect the linear term to be positive
    assert features["coefficients"][0] > 0

    # We would also expect this for the quadratic and cubic terms
    assert features["coefficients"][1] > 0
    assert features["coefficients"][2] > 0

    # Verify that melodies with a single note have 0.0 coefficients
    # for all 3 terms
    single_note = Score.from_melody(pitches=[60], durations=[1.0])
    single_note_features = fantastic_polynomial_contour_features(single_note)
    assert single_note_features["coefficients"] == [0.0, 0.0, 0.0]

    # Produce a melody with an arching contour
    arching = Score.from_melody(
        pitches=[60, 62, 64, 65, 67, 72, 67, 65, 64, 62, 60],
        durations=[1.0] * 11,
    )
    arching_features = fantastic_polynomial_contour_features(arching)

    # We expect that the quadratic term is greater in magnitude than the linear term
    assert abs(arching_features["coefficients"][1]) > abs(
        arching_features["coefficients"][0]
    )


def test_fantastic_huron_contour_features():
    melody = Score.from_melody(
        pitches=[60, 62, 64, 65, 67, 72], durations=[1.0] * 6
    )
    features = fantastic_huron_contour_features(melody)
    assert features is not None

    # The first pitch should be 60
    assert features["first_pitch"] == 60
    # The mean pitch should be 67
    assert features["mean_pitch"] == 67
    # The last pitch should be 72
    assert features["last_pitch"] == 72
    # The first to mean difference should be 7
    assert features["first_to_mean"] == 7
    # The mean to last difference should be 5
    assert features["mean_to_last"] == 5
    # The contour class should be "Ascending-Ascending"
    assert features["contour_class"] == "Ascending-Ascending"


def test_fantastic_mtype_summary_features():
    melody = Score.from_melody(
        pitches=[56, 58, 61, 58, 65, 65, 63],
        durations=[0.25, 0.25, 0.25, 0.25, 0.75, 0.75, 0.5],
    )

    # Test with no segmentation first
    features = fantastic_mtype_summary_features(
        melody, segment=False, phrase_gap=1.0, units="quarters"
    )

    # Verify all summary stats are present and have reasonable values
    assert 0 <= features["mean_entropy"] <= 1.0
    assert 0 <= features["mean_productivity"] <= 1.0
    assert features["yules_k"] >= 0
    assert 0 <= features["simpsons_d"] <= 1.0
    assert 0 <= features["sichels_s"] <= 1.0

    # Honore's H should be incalculable for this melody
    assert math.isnan(features["honores_h"])

    # Test with a longer melody that can be segmented
    long_melody = Score.from_melody(
        pitches=[
            60,
            62,
            64,
            60,  # First phrase
            67,
            67,
            65,
            64,  # Second phrase
            62,
            62,
            60,
            60,  # Third phrase
            64,
            64,
            62,
            60,  # Fourth phrase
        ],
        durations=[
            1.0,
            1.0,
            1.0,
            2.0,  # First phrase with longer final note
            1.0,
            1.0,
            1.0,
            2.0,  # Second phrase
            1.0,
            1.0,
            1.0,
            2.0,  # Third phrase
            1.0,
            1.0,
            1.0,
            2.0,  # Fourth phrase
        ],
    )

    # Test with segmentation
    segmented_features = fantastic_mtype_summary_features(
        long_melody, segment=True, phrase_gap=1.5, units="quarters"
    )

    # Features should still be in valid ranges
    assert 0 <= segmented_features["mean_entropy"] <= 1.0
    assert 0 <= segmented_features["mean_productivity"] <= 1.0
    assert segmented_features["yules_k"] >= 0
    assert 0 <= segmented_features["simpsons_d"] <= 1.0
    assert 0 <= segmented_features["sichels_s"] <= 1.0

    # Honore's H is calculable for this melody
    assert segmented_features["honores_h"] >= 0


def test_fantastic_pitch_features():
    melody = Score.from_melody(
        pitches=[56, 58, 61, 58, 65, 65, 63],
        durations=[0.25, 0.25, 0.25, 0.25, 0.75, 0.75, 0.5],
    )

    features = fantastic_pitch_features(melody)
    assert features is not None

    # The melody has a range of 9
    assert features["pitch_range"] == 9
    # The standard deviation should be positive and reasonable given the pitch range
    assert features["pitch_std"] > 0
    assert features["pitch_std"] < features["pitch_range"]

    # Entropy should be between 0 and 1 since it's normalized by log2(24)
    assert 0 <= features["pitch_entropy"] <= 1.0

    # Test with a melody that has all the same pitch
    flat_melody = Score.from_melody(
        pitches=[60, 60, 60, 60, 60], durations=[1.0, 1.0, 1.0, 1.0, 1.0]
    )
    flat_features = fantastic_pitch_features(flat_melody)

    # A melody with all same pitches should have:
    assert flat_features["pitch_range"] == 0  # No range
    assert flat_features["pitch_std"] == 0  # No variation
    assert flat_features["pitch_entropy"] == 0  # Minimum entropy

    # Test with a chromatic scale
    chromatic = Score.from_melody(
        pitches=[60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71],
        durations=[1.0] * 12,
    )
    chrom_features = fantastic_pitch_features(chromatic)

    # Chromatic scale should have:
    assert chrom_features["pitch_range"] == 11  # Range of 11 semitones
    assert chrom_features["pitch_std"] > 0  # Some variation
    # Relatively high entropy since all pitches occur equally often
    assert chrom_features["pitch_entropy"] > 0.5


def test_fantastic_pitch_interval_features():
    melody = Score.from_melody(
        pitches=[56, 58, 61, 58, 65, 65, 63],
        durations=[0.25, 0.25, 0.25, 0.25, 0.75, 0.75, 0.5],
    )

    features = fantastic_pitch_interval_features(melody)
    assert features is not None

    # The intervals in the melody are: [2, 3, -3, 7, 0, -2]
    # The absolute intervals are: [2, 3, 3, 7, 0, 2]
    assert features["absolute_interval_range"] == 7  # max(7) - min(0) = 7
    assert features["mean_absolute_interval"] == pytest.approx(2.833, rel=0.01)
    assert features["std_absolute_interval"] > 0
    assert features["modal_interval"] == 2  # 2 appears twice
    assert 0 <= features["interval_entropy"] <= 1.0  # Normalized by log2(23)

    # Test with a melody that has all the same intervals
    repeated_melody = Score.from_melody(
        pitches=[60, 62, 64, 66, 68, 70], durations=[1.0] * 6
    )
    repeated_features = fantastic_pitch_interval_features(repeated_melody)

    # All intervals are 2 semitones
    assert (
        repeated_features["absolute_interval_range"] == 0
    )  # max(2) - min(2) = 0
    assert repeated_features["mean_absolute_interval"] == 2.0
    assert repeated_features["std_absolute_interval"] == 0
    assert repeated_features["modal_interval"] == 2
    assert repeated_features["interval_entropy"] == 0  # Minimum entropy

    # Test with alternating up/down intervals
    zigzag = Score.from_melody(
        pitches=[60, 65, 60, 65, 60, 65], durations=[1.0] * 6
    )
    zigzag_features = fantastic_pitch_interval_features(zigzag)

    # All intervals are +5 or -5
    assert (
        zigzag_features["absolute_interval_range"] == 0
    )  # max(5) - min(5) = 0
    assert zigzag_features["mean_absolute_interval"] == 5.0
    assert zigzag_features["std_absolute_interval"] == 0
    assert zigzag_features["modal_interval"] == 5
    assert zigzag_features["interval_entropy"] == 0  # Only one interval size



"""
An Efficient Algorithm for Computing an Inner Metric Analysis (IMA) of Music
Created by Brian Bemman based on algorithms provided in two forthcoming papers, Bemman, B. et al. (forthcoming a,b)

Computes an Inner Metric Analysis of music (Volk, 2008), which consists of the following two steps:

1. enumerating all local meters in a strictly increasing sequence of integer onsets; and
2. computing the metric and spectral weight profiles from all enumerated local meters.

- A local meter is a maximal arithmetic progression of at least l plus 1 onset pulses that is not contained in any
other. The input parameter, l, specifies the minimum local meter length (default 2 i.e., three onset pulses).

- A metric weight profile consists of a weight for each onset, l'^p, summed over all local meter pulses coinciding at
this onset, where l' is the length (minus 1) of a given local meter and p is an input weighting parameter (default 2).

- A spectral weight profile consists of the same weight, l'^p, computed for each time point in an underlying metrical
grid and summed over all local meter pulses and extensions coinciding at this time point.

Example:

    onsets: [0, 1, 2, 3, 4, 5, 7]
    minimum local meter length: l = 2
    weighting parameter: p = 2

    local meters: [0, 1, 2, 3, 4, 5], [1, 3, 5, 7], [1, 4, 7]
    metric weights: [25, 38, 25, 34, 29, 34, 0, 13]
    spectral weights: [25, 38, 25, 34, 29, 34, 25, 38]

Usage:

    Compute an Inner Metric Analysis with a minimum local meter length, l = 2, and weighting parameter, p = 2, and
    return the metric and spectral weight profiles but not the local meters found (if any).

    analysis = compute_inner_metric_analysis(onsets, 2, 2, False)

    print(analysis.metric_weights)
    print(analysis.spectral_weights)
    print(analysis.local_meters)

    'analysis.local_meters' will be non-empty only if return_local_meters is True and local meters have been found. Each
    local meter is encoded as a tuple, (start_onset, period, length). For example, the local meter, [1, 3, 5, 7], above
    is (start_onset=1, period=2, length=3).

A note on performance:

For the vast majority of individual musical works (i.e., those having fewer than 5,000 onsets), one can expect a running
time not exceeding a few seconds. For significantly larger works, such as Richard Wagner's Das Rheingold, having roughly
32,000 onsets, one can expect a running time of a few minutes. Should these times prove inadequate for your application
or larger sequences need to be run, a considerably more performant C++ version is available at
https://github.com/brianm2b. In either case, depending on your machine, memory usage may become prohibitive for
sequences of any significantly greater length.

<small>**Author**: Brian Bemman</small>
"""

__author__ = "Brian Bemman"

from array import array
from typing import NamedTuple


class LM(NamedTuple):
    """Small class for storing a local meter (LM)."""

    start_onset: int
    period: int
    length: int  # number of pulses minus 1


class IMA(NamedTuple):
    """Small class for storing an Inner Metric Analysis (IMA)."""

    metric_weights: array
    spectral_weights: array
    local_meters: list[LM]


def check_input_is_well_formed(
    onsets: list, l_param: int, p_weight_param: int
) -> list[int]:
    """Checks for well-formed input parameters and preprocesses the input sequence of onsets.

    Parameters:
    - onsets: strictly increasing sequence of non-negative integer onsets
    - l_param: minimum local meter length, l (default 2)
    - p_weight_param: weighting parameter, p, for controlling the influence of each local meter (default 2)

    Returns:
    - onsets: strictly increasing sequence of non-negative integer onsets
    """
    n = len(onsets)
    if n < 3:
        raise ValueError(
            "Number of onsets is less than three, so no local meters possible"
        )
    if not (2 <= l_param < n):
        raise ValueError(
            "Minimum local meter length, l, must be 2 <= l < len(onsets). Default is l = 2"
        )
    if p_weight_param not in (0, 1, 2):
        raise ValueError(
            "Weighting parameter, p, must be equal to 0, 1, or 2. Default is p = 2"
        )
    prev = None
    processed_onsets = []
    for onset in onsets:
        if not (
            isinstance(onset, int)
            or (isinstance(onset, float) and onset.is_integer())
        ):
            raise TypeError("Onsets must be whole numbers")
        onset = int(onset)
        if onset < 0:
            raise ValueError("Onsets must be non-negative")
        if prev is not None and onset <= prev:
            raise ValueError("Onsets must be strictly increasing")
        processed_onsets.append(onset)
        prev = onset
    return processed_onsets


def pairwise_arithmetic_progression_lengths(
    onsets: list[int],
) -> tuple[array, list[int], array, int, list[int]]:
    """Computes maximal lengths of all (i < j) arithmetic progressions using dynamic programming (Erickson 1999).

    Parameters:
    - onsets: strictly increasing sequence of non-negative integer onsets

    Returns:
    - ap_length_table: upper-right triangle of table containing lengths of all (i<j) arithmetic progressions in onsets
    - discovered_periods: all distinct periods belonging to all arithmetic progression triples found in onsets
    - max_period: largest discovered period from all arithmetic progression triples found in onsets

    Note that this is a modification of an algorithm by Erickson (1999) for computing the longest arithmetic progression
    """
    n = len(onsets)
    max_onset = onsets[-1]
    min_onset = onsets[0]
    discovered_periods: list[int] = []
    max_period = 0
    ap_length_table = array("i", [2]) * ((n * (n - 1)) // 2)
    period_indices = array("i", [-1]) * (((max_onset - min_onset) // 2) + 1)
    row_base = [0] * n
    for i in range(0, n):
        row_base[i] = i * (n - 1) - (i * (i - 1)) // 2

    for j in range(n - 2, 0, -1):  # j from n-2 down to 1 (inclusive)
        i = j - 1
        k = j + 1
        second_term_sum = 2 * onsets[j]
        row_base_j = row_base[j]
        while i >= 0 and k < n:
            first_third_term_sum = onsets[i] + onsets[k]
            if (
                first_third_term_sum < second_term_sum
            ):  # right-side term too small
                k += 1
                continue
            if (
                first_third_term_sum > second_term_sum
            ):  # left-side term too large
                i -= 1
                continue
            # arithmetic triple
            idx_ij = row_base[i] + (j - i - 1)  # linear index of i, j
            idx_jk = row_base_j + (k - j - 1)  # linear index of j, k
            prev_len = ap_length_table[idx_jk]
            updated_len = (prev_len + 1) if (prev_len > 2) else 3
            ap_length_table[idx_ij] = updated_len
            ap_period = onsets[j] - onsets[i]
            if period_indices[ap_period] < 0:  # new period discovered
                period_indices[ap_period] = len(discovered_periods)
                discovered_periods.append(ap_period)
                if ap_period > max_period:
                    max_period = ap_period
            i -= 1
            k += 1
    return (
        ap_length_table,
        discovered_periods,
        period_indices,
        max_period,
        row_base,
    )


def compute_period_cofactors(
    discovered_periods, period_indices, max_period
) -> tuple[list[int], list[int]]:
    """Computes for each discovered period all cofactors from division by its prime factors.

    Parameters:
    - discovered_periods: all distinct periods belonging to all arithmetic progression triples found in onsets
    - period_indices: index in discovered_periods associated to the value of its period
    - max_period: largest discovered period from all arithmetic progression triples found in onsets

    Returns:
    - period_factors: all cofactors from division by its prime factors ordered by discovered_periods
    - factor_indices: starting indices of cofactors in period_factors ordered by discovered_periods (i.e., prefix sum)

    Note that this procedure makes use of Euler's Linear Sieve algorithm.
    """
    smallest_prime_factors = array("i", [0]) * (max_period + 1)  # Euler sieve
    next_distinct = array("i", [0]) * (
        max_period + 1
    )  # index of next distinct prime factor
    primes = []
    for i in range(2, max_period + 1):  # Euler's Linear Sieve procedure
        if smallest_prime_factors[i] == 0:
            smallest_prime_factors[i] = i
            primes.append(i)
            next_distinct[i] = 1
        for p in primes:
            mul = i * p  # multiple of prime
            if mul > max_period:
                break
            smallest_prime_factors[mul] = p
            if p == smallest_prime_factors[i]:
                next_distinct[mul] = next_distinct[i]
                break
            else:
                next_distinct[mul] = i
    period_factors: list[int] = (
        []
    )  # cofactors for each period in discovered_periods (same order)
    factor_indices: list[int] = [
        0
    ]  # index of the first cofactor for each period in discovered_periods
    for (
        period
    ) in (
        discovered_periods
    ):  # compute period cofactors from each distinct prime factor
        idx = period
        while idx > 1:
            cofactor = (
                period // smallest_prime_factors[idx]
            )  # p-free part of period
            if period_indices[cofactor] > -1:
                period_factors.append(cofactor)
            idx = next_distinct[
                idx
            ]  # jump to index of next distinct prime factor
        factor_indices.append(len(period_factors))
    return period_factors, factor_indices


def is_local_meter(
    factor_start_idx: int,
    factor_end_idx: int,
    period_factors: list[int],
    row_periods: list[tuple[int, int]],
    current_row: int,
    ap_start_onset: int,
    ap_end_onset: int,
    period_indices: array,
) -> bool:
    """Determines if a maximal arithmetic progression is a local meter.

    Parameters
    ----------
    - factor_start_idx: index of first factor in period_factors for the period of a maximal arithmetic progressions
    - factor_end_idx: index of first factor in period_factors for the next (i.e., i + 1) period in period_factors
    - period_factors: for the period of a maximal arithmetic progression, all factors from division by its prime factors
    - row_periods: maximal length of each (i < j) arithmetic progression from a given i associated to its period
    - current_row: i-th row in ap_length_table
    - ap_start_onset: first onset of a maximal arithmetic progression
    - ap_end_onset: last onset of a maximal arithmetic progression
    - period_indices: index in discovered_periods associated to the value of its period

    Returns
    -------
    - boolean: indicating whether a maximal arithmetic progression is a local meter
    """
    for idx in range(
        factor_start_idx, factor_end_idx
    ):  # cofactors for ap_period
        cofactor = period_factors[idx]
        cofactor_length, cofactor_row = row_periods[period_indices[cofactor]]
        if (
            cofactor_row == current_row
            and ap_end_onset <= ap_start_onset + cofactor * cofactor_length
        ):
            return False  # existing ap with a cofactor of ap_period at least as long
    return True


def update_weights(
    metric_weights: array,
    spectral_weights: array,
    ap_start_onset: int,
    ap_end_onset: int,
    ap_period: int,
    ap_length: int,
    max_onset: int,
    min_onset: int,
    p_weight_param: int,
):
    """Updates the metric and spectral weight profiles of a musical piece for a given local meter.

    Parameters
    ----------
    - metric_weights: sequence of metric weights, one for each onset of a local meter
    - spectral_weights: sequence of spectral weights, one for each time point of a local meter or extension
    - ap_start_onset: first onset of a local meter
    - ap_end_onset: last onset of a local meter
    - ap_period: period of a local meter
    - ap_length: length minus 1 of a local meter
    - max_onset: greatest onset in onsets
    - min_onset: smallest onset in onsets
    - p_weight_param: weighting parameter, p, for controlling the influence of each local meter (default 2)

    Returns
    -------
    - metric_weights: sequence of metric weights, one for each onset of a local meter
    - spectral_weights: sequence of spectral weights, one for each time point of a local meter or extension
    """
    weight = ap_length**p_weight_param  # l^p weight for this local meter
    time_point = ap_start_onset  # time point in underlying grid
    while time_point <= ap_end_onset:  # local meter (pulses/onsets)
        metric_weights[time_point] += weight
        spectral_weights[time_point] += weight
        time_point += ap_period
    while time_point <= max_onset:  # right side of local meter
        spectral_weights[time_point] += weight
        time_point += ap_period
    time_point = ap_start_onset - ap_period
    while time_point >= min_onset:  # left side of local meter
        spectral_weights[time_point] += weight
        time_point -= ap_period
    return metric_weights, spectral_weights


def compute_inner_metric_analysis(
    onsets: list,
    l_param: int = 2,
    p_weight_param: int = 2,
    return_local_meters: bool = False,
) -> IMA:
    """Computes an Inner Metric Analysis (IMA) of music (Volk 2008).

    Parameters
    ----------
    - onsets: strictly increasing sequence of non-negative integer onsets
    - l_param: minimum local meter length, l (default 2)
    - p_weight_param: weighting parameter, p, for controlling the influence of each local meter (default 2)
    - return_local_meters: boolean indicating whether to store and return all discovered local meters (default False)

    Returns
    -------
    - An IMA object consisting of the following three attributes:
        - metric_weights: sequence of metric weights, one for each onset of a local meter
        - spectral_weights: sequence of spectral weights, one for each time point of a local meter or extension
        - local_meters: all local meters of the form LM(ap_start_onset, period, ap_length) discovered in onsets

    Note that local_meters will be non-empty if 'return_local_meters' is True and local meters have been found

    Example
    -------
    Volk (2008) provides an example with 19 local meters:
    >>> onsets = [0, 1, 2, 6, 8, 9, 10, 14, 16, 17, 18, 22, 24, 25, 26, 30]
    >>> analysis = compute_inner_metric_analysis(onsets, 2, 2, True)
    >>> analysis.metric_weights
    array('i', [17, 13, 65, 0, 0, 0, 57, 0, 25, 21, 65, 0, 0, 0, 57, 0, 33, 21, 65, 0, 0, 0, 57, 0, 25, 13, 65, 0, 0, 0, 57])

    >>> analysis.spectral_weights
    array('i', [45, 25, 89, 20, 32, 16, 85, 16, 45, 33, 85, 16, 32, 16, 81, 16, 53, 33, 85, 16, 32, 16, 81, 20, 45, 25, 89, 20, 32, 16, 89])

    >>> len(analysis.local_meters)
    19

    """
    # check whether onsets are properly formed
    onsets = check_input_is_well_formed(onsets, l_param, p_weight_param)

    # compute maximal lengths of all (i < j) arithmetic progressions
    (
        ap_length_table,
        discovered_periods,
        period_indices,
        max_period,
        row_base,
    ) = pairwise_arithmetic_progression_lengths(onsets)
    # compute (prime factor | period) cofactors for all discovered periods in onsets
    period_factors, factor_indices = compute_period_cofactors(
        discovered_periods, period_indices, max_period
    )

    n = len(onsets)
    max_onset = onsets[-1]
    min_onset = onsets[0]
    metric_weights = array("i", [0]) * (max_onset + 1)
    spectral_weights = array("i", [0]) * (max_onset + 1)
    local_meters: list[LM] = []
    row_periods = [
        (0, -1) for _ in range(len(discovered_periods))
    ]  # (ap length, current row) for each period
    indicator_onsets = set(onsets)

    for i in range(n - 2):  # each row in ap_length_table
        ap_start_onset = onsets[i]
        max_period_in_row = min(
            (max_onset - ap_start_onset) // l_param, max_period
        )  # maximum possible period
        row_start_idx = row_base[i]
        base_offset = row_start_idx - i - 1  # constant offset for row i
        for j in range(i + 1, n - 1):
            ap_period = onsets[j] - ap_start_onset
            if (
                ap_period > max_period_in_row
            ):  # remaining j ap lengths for this i will be of length 2
                break
            current_idx = base_offset + j  # flat index for (i < j)
            ap_length = ap_length_table[current_idx] - 1  # length minus 1
            if ap_length >= l_param:  # at least length l_param
                period_idx = period_indices[
                    ap_period
                ]  # index in discovered_periods/factor_indices/period_factors
                row_periods[period_idx] = (ap_length, i)
                ap_prev_onset = (
                    ap_start_onset - ap_period
                )  # prior possible onset
                if (
                    ap_prev_onset < 0 or ap_prev_onset not in indicator_onsets
                ):  # new maximal arithmetic progression
                    ap_end_onset = ap_start_onset + ap_period * ap_length
                    factor_start_idx = factor_indices[
                        period_idx
                    ]  # index of first factor of ap_period
                    factor_end_idx = factor_indices[
                        period_idx + 1
                    ]  # index of first factor of next period
                    if is_local_meter(
                        factor_start_idx,
                        factor_end_idx,
                        period_factors,
                        row_periods,
                        i,
                        ap_start_onset,
                        ap_end_onset,
                        period_indices,
                    ):  # local meter found
                        # update metric and spectral weight profiles
                        metric_weights, spectral_weights = update_weights(
                            metric_weights,
                            spectral_weights,
                            ap_start_onset,
                            ap_end_onset,
                            ap_period,
                            ap_length,
                            max_onset,
                            min_onset,
                            p_weight_param,
                        )
                        if return_local_meters:
                            local_meters.append(
                                LM(ap_start_onset, ap_period, ap_length)
                            )
    return IMA(metric_weights, spectral_weights, local_meters)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Test suite for tempo modelling functions inside amads/time/tempo.py"""

import numpy as np
import pytest

from amads.time.tempo import (
    _validate_beats,
    beats_to_tempo,
    tempo_fluctuation,
    tempo_mean,
    tempo_slope,
)

# Quarter note beat positions for different "performances"
STABLE_PERFORMANCE = [1.0, 2.0, 3.0, 4.0]
ACCELERATING_PERFORMANCE = [1.0, 1.9, 2.7, 3.4]
DECELERATING_PERFORMANCE = [1.0, 2.1, 3.3, 4.6]


def test_mean_tempo():
    stable_bpm_mean = 60.0
    assert pytest.approx(tempo_mean(STABLE_PERFORMANCE)) == stable_bpm_mean
    accelerating_bpm_mean = 75.7936
    assert (
        pytest.approx(tempo_mean(ACCELERATING_PERFORMANCE), abs=1e-4)
        == accelerating_bpm_mean
    )
    decelerating_bpm_mean = 50.2331
    assert (
        pytest.approx(tempo_mean(DECELERATING_PERFORMANCE), abs=1e-4)
        == decelerating_bpm_mean
    )


def test_tempo_slope():
    assert tempo_slope(STABLE_PERFORMANCE) == 0.0
    assert tempo_slope(ACCELERATING_PERFORMANCE) > 0
    assert tempo_slope(DECELERATING_PERFORMANCE) < 0


def test_tempo_fluctuation():
    # Stable performance should not fluctuate
    assert tempo_fluctuation(STABLE_PERFORMANCE) == 0.0
    # Unstable performances should fluctuate
    assert tempo_fluctuation(ACCELERATING_PERFORMANCE) > 0
    assert tempo_fluctuation(DECELERATING_PERFORMANCE) > 0


def test_beats_to_tempo():
    # With a stable performance with IBI = 1.0, we'll expect a full array of 60 BPMs
    stable_bpms = beats_to_tempo(STABLE_PERFORMANCE)
    assert np.array_equal(
        stable_bpms.round(1),
        np.array(
            [
                60.0,
                60.0,
                60.0,
            ]
        ),
    )
    # Test another performance that accelerates: IBIs decrease, BPMs increase
    accelerating_bpms = beats_to_tempo(ACCELERATING_PERFORMANCE)
    assert np.array_equal(
        accelerating_bpms.round(1), np.array([66.7, 75, 85.7])
    )
    # Test another performance that decelerates: IBIs increase, BPMs decrease
    decelerating_bpms = beats_to_tempo(DECELERATING_PERFORMANCE)
    assert np.array_equal(
        decelerating_bpms.round(1), np.array([54.5, 50.0, 46.2])
    )


def test_validate_beats():
    # Valid case (should not raise an error)
    valid_beats = [0.5, 1.0, 1.5, 2.0, 2.5]
    assert _validate_beats(valid_beats) is None
    # Too few beats
    with pytest.raises(ValueError):
        _validate_beats([0.5])
    # Non-1D array
    with pytest.raises(ValueError):
        _validate_beats([[0.5, 1.0], [1.5, 2.0]])
    # Unsorted beats
    with pytest.raises(ValueError):
        _validate_beats([0.5, 1.5, 1.0, 2.0])
    # Duplicate values
    with pytest.raises(ValueError):
        _validate_beats([0.5, 1.0, 1.0, 2.0])  # Duplicate 1.0
    # NaN values
    with pytest.raises(ValueError):
        _validate_beats([0.5, 1.0, None, 2.0])
